#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


from typing import Callable, Dict, List, Optional
from ads.common import logger
from ads.model.extractor.automl_extractor import AutoMLExtractor
from ads.model.generic_model import FrameworkSpecificModel
from ads.model.model_properties import ModelProperties
from ads.model.serde.common import SERDE
from ads.common.decorator.deprecate import deprecated


class AutoMLModel(FrameworkSpecificModel):
    """AutoMLModel class for estimators from AutoML framework.

    Attributes
    ----------
    algorithm: str
        "ensemble", the algorithm name of the model.
    artifact_dir: str
        Artifact directory to store the files needed for deployment.
    auth: Dict
        Default authentication is set using the `ads.set_auth` API. To override the
        default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create
        an authentication signer to instantiate an IdentityClient object.
    estimator: Callable
        A trained automl estimator/model using oracle automl.
    framework: str
        "oracle_automl", the framework name of the estimator.
    hyperparameter: dict
        The hyperparameters of the estimator.
    metadata_custom: ModelCustomMetadata
        The model custom metadata.
    metadata_provenance: ModelProvenanceMetadata
        The model provenance metadata.
    metadata_taxonomy: ModelTaxonomyMetadata
        The model taxonomy metadata.
    model_artifact: ModelArtifact
        This is built by calling prepare.
    model_deployment: ModelDeployment
        A ModelDeployment instance.
    model_file_name: str
        Name of the serialized model. Default to "model.pkl".
    model_id: str
        The model ID.
    properties: ModelProperties
        ModelProperties object required to save and deploy model.
        For more details, check https://accelerated-data-science.readthedocs.io/en/latest/ads.model.html#module-ads.model.model_properties.
    runtime_info: RuntimeInfo
        A RuntimeInfo instance.
    schema_input: Schema
        Schema describes the structure of the input data.
    schema_output: Schema
        Schema describes the structure of the output data.
    serialize: bool
        Whether to serialize the model to pkl file by default. If False, you need to serialize the model manually,
        save it under artifact_dir and update the score.py manually.
    version: str
        The framework version of the model.

    Methods
    -------
    delete_deployment(...)
        Deletes the current model deployment.
    deploy(..., **kwargs)
        Deploys a model.
    from_model_artifact(uri, model_file_name, artifact_dir, ..., **kwargs)
        Loads model from the specified folder, or zip/tar archive.
    from_model_catalog(model_id, model_file_name, artifact_dir, ..., **kwargs)
        Loads model from model catalog.
    introspect(...)
        Runs model introspection.
    predict(data, ...)
        Returns prediction of input data run against the model deployment endpoint.
    prepare(..., **kwargs)
        Prepare and save the score.py, serialized model and runtime.yaml file.
    reload(...)
        Reloads the model artifact files: `score.py` and the `runtime.yaml`.
    save(..., **kwargs)
        Saves model artifacts to the model catalog.
    summary_status(...)
        Gets a summary table of the current status.
    verify(data, ...)
        Tests if deployment works in local environment.

    Examples
    --------
    >>> import tempfile
    >>> import logging
    >>> import warnings
    >>> from ads.automl.driver import AutoML
    >>> from ads.automl.provider import OracleAutoMLProvider
    >>> from ads.dataset.dataset_browser import DatasetBrowser
    >>> from ads.model.framework.automl_model import AutoMLModel
    >>> from ads.model.model_metadata import UseCaseType
    >>> ds = DatasetBrowser.sklearn().open("wine").set_target("target")
    >>> train, test = ds.train_test_split(test_size=0.1, random_state = 42)

    >>> ml_engine = OracleAutoMLProvider(n_jobs=-1, loglevel=logging.ERROR)
    >>> oracle_automl = AutoML(train, provider=ml_engine)
    >>> model, baseline = oracle_automl.train(
    ...                model_list=['LogisticRegression', 'DecisionTreeClassifier'],
    ...                random_state = 42,
    ...                time_budget = 500
    ...        )

    >>> automl_model.prepare(inference_conda_env=inference_conda_env, force_overwrite=True)
    >>> automl_model.verify(...)
    >>> automl_model.save()
    >>> model_deployment = automl_model.deploy(wait_for_completion=False)
    """

    _PREFIX = "automl"

    @deprecated(
        details=f"Working with AutoML has moved from within ADS to working directly with the automlx library. Please use `GenericModel` https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/frameworks/genericmodel.html class instead. An example can be found at https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/frameworks/automlmodel.html",
        raise_error=True,
    )
    def __init__(
        self,
        estimator: Callable,
        artifact_dir: str,
        properties: Optional[ModelProperties] = None,
        auth: Dict = None,
        model_save_serializer: Optional[SERDE] = None,
        model_input_serializer: Optional[SERDE] = None,
        **kwargs,
    ):
        """
        Initiates a AutoMLModel instance.

        Parameters
        ----------
        estimator: Callable
            Any model object generated by automl framework.
        artifact_dir: str
            Directory for generate artifact.
        properties: (ModelProperties, optional). Defaults to None.
            ModelProperties object required to save and deploy model.
        auth: (Dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.
        model_save_serializer: (SERDE or str, optional). Defaults to None.
            Instance of ads.model.SERDE. Used for serialize/deserialize model.
        model_input_serializer: (SERDE, optional). Defaults to None.
            Instance of ads.model.SERDE. Used for serialize/deserialize data.

        Returns
        -------
        AutoMLModel
            AutoMLModel instance.

        Raises
        ------
        TypeError
            If the input model is not an AutoML model.
        """
        if not str(type(estimator)).startswith("<class 'ads.common.model.ADSModel'"):
            raise TypeError(f"{str(type(estimator))} is not supported in AutoMLModel.")
        super().__init__(
            estimator=estimator,
            artifact_dir=artifact_dir,
            properties=properties,
            auth=auth,
            model_save_serializer=model_save_serializer,
            model_input_serializer=model_input_serializer,
            **kwargs,
        )
        self._extractor = AutoMLExtractor(estimator)
        self.framework = self._extractor.framework
        self.algorithm = self._extractor.algorithm
        self.version = self._extractor.version
        self.hyperparameter = self._extractor.hyperparameter



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
from ads import __version__
from enum import Enum, auto

import ads.config

EXTRA_USER_AGENT_INFO = "EXTRA_USER_AGENT_INFO"
USER_AGENT_KEY = "additional_user_agent"
ENV_MD_OCID = "MD_OCID"
UNKNOWN = "UNKNOWN"


class Surface(Enum):
    """
    An Enum class for labeling the surface where ADS is being used
    """

    WORKSTATION = auto()
    DATASCIENCE_JOB = auto()
    DATASCIENCE_NOTEBOOK = auto()
    DATASCIENCE_MODEL_DEPLOYMENT = auto()
    DATAFLOW = auto()
    OCI_SERVICE = auto()

    @classmethod
    def surface(cls):
        surface = cls.WORKSTATION
        if (
            ads.config.OCI_RESOURCE_PRINCIPAL_VERSION
            or ads.config.OCI_RESOURCE_PRINCIPAL_RPT_PATH
            or ads.config.OCI_RESOURCE_PRINCIPAL_RPT_ID
        ):
            surface = cls.OCI_SERVICE
            if ads.config.JOB_RUN_OCID:
                surface = cls.DATASCIENCE_JOB
            elif ads.config.NB_SESSION_OCID:
                surface = cls.DATASCIENCE_NOTEBOOK
            elif ads.config.MD_OCID:
                surface = cls.DATASCIENCE_MODEL_DEPLOYMENT
            elif ads.config.DATAFLOW_RUN_OCID:
                surface = cls.DATAFLOW
        return surface


def update_oci_client_config(config={}):
    if not config.get(USER_AGENT_KEY):
        config[
            USER_AGENT_KEY
        ] = f"Oracle-ads/version={__version__}/surface={Surface.surface().name}"  # To be enabled in future - /api={os.environ.get(EXTRA_USER_AGENT_INFO,UNKNOWN)}"
    return config



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import sys
import os
import json
from setuptools import setup, find_packages
from functools import reduce
from pathlib import Path
from setuptools.command.install import install
from setuptools.command.develop import develop


install_requires = [
    "asteval>=0.9.25",
    "cerberus>=1.3.4",
    "cloudpickle>=1.6.0",
    "fsspec>=0.8.7",
    "jinja2>=2.11.2",
    "gitpython>=3.1.2",
    "matplotlib>=3.1.3",
    "numpy>=1.19.2",
    "oci>=2.93.1",
    "ocifs>=1.1.3",
    "pandas>1.2.1,<1.6",
    "python_jsonschema_objects>=0.3.13",
    "PyYAML>=5.4,<6",
    "requests",
    "scikit-learn>=0.23.2,<1.2",
    "tabulate>=0.8.9",
    "tqdm>=4.59.0",
    "psutil>=5.7.2",
]

extras_require = {
    "boosted": [
        "xgboost",
        "lightgbm",
    ],
    "notebook": [
        "ipywidgets~=7.6.3",
        "ipython>=7.23.1, <8.0",
    ],
    "text": ["wordcloud>=1.8.1", "spacy"],
    "viz": [
        "bokeh>=2.3.0, <=2.4.3",
        "folium>=0.12.1",
        "graphviz<0.17",
        "scipy>=1.5.4",
        "seaborn>=0.11.0",
    ],
    "data": [
        "fastavro>=0.24.2",
        "openpyxl>=3.0.7",
        "pandavro>=1.6.0",
        "datefinder>=0.7.1",
        "htmllistparse>=0.6.0",
        "sqlalchemy>=1.4.1, <=1.4.46",
        "oracledb>=1.0",
    ],
    "opctl": [
        "oci-cli",
        "docker",
        "conda-pack",
        "nbconvert",
        "nbformat",
        "inflection",
    ],
    "mysql": ["mysql-connector-python"],
    "bds": ["ibis-framework[impala]", "hdfs[kerberos]", "sqlalchemy"],
    "spark": ["pyspark>=3.0.0"],
    "huggingface": ["transformers"],
}

this_directory = Path(__file__).parent


def update_extra_with_internal_packages():
    loaded_dep = {}
    internal_deps = os.path.join(this_directory, "internal_extra_dependency.json")
    print(f"looking for {internal_deps}")
    if os.path.exists(internal_deps):
        with open(internal_deps) as idf:
            loaded_dep = json.load(idf)
            print(f"Found: {loaded_dep}")
    return loaded_dep


extras_require.update(update_extra_with_internal_packages())

extras_require["torch"] = extras_require["viz"] + ["torch"] + ["torchvision"]
extras_require["tensorflow"] = extras_require["viz"] + [
    "tensorflow",
]
extras_require["geo"] = extras_require["viz"] + ["geopandas"]
extras_require["onnx"] = extras_require["viz"] + [
    "protobuf<=3.20",
    "onnx>=1.12.0",
    "onnxruntime>=1.10.0",
    "onnxmltools>=1.10.0",
    "skl2onnx>=1.10.4",
    "tf2onnx",
    "xgboost==1.5.1",
    "lightgbm==3.3.1",
]
extras_require["optuna"] = extras_require["viz"] + ["optuna==2.9.0"]

extras_require["complete"] = sorted({v for req in extras_require.values() for v in req})
extras_require["all-optional"] = reduce(
    list.__add__,
    [
        extras_require[k]
        for k in extras_require
        if k not in ["boosted", "opctl", "complete"]
    ],
)
extras_require["all-public"] = reduce(
    list.__add__,
    [
        extras_require[k]
        for k in extras_require
        if k not in ["all-optional", "complete"]
    ],
)
# Only include pytest-runner in setup_requires if we're invoking tests
if {"pytest", "test", "ptr"}.intersection(sys.argv):
    setup_requires = ["pytest-runner"]
else:
    setup_requires = []

ADS_VERSION = "UNKNOWN"
with open(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), "ads", "ads_version.json")
) as version_file:
    ADS_VERSION = json.load(version_file)["version"]


long_description = (this_directory / "README.md").read_text()
setup(
    name="oracle_ads",
    version=ADS_VERSION,
    description="Oracle Accelerated Data Science SDK",
    author="Oracle Data Science",
    license="Universal Permissive License 1.0",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    url="https://docs.oracle.com/en-us/iaas/tools/ads-sdk/latest/index.html",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Universal Permissive License (UPL)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    keywords="Oracle Cloud Infrastructure, OCI, Machine Learning, ML, Artificial Intelligence, AI, Data Science, Cloud, Oracle",
    include_package_data=True,
    install_requires=install_requires,
    python_requires=">=3.8",
    setup_requires=setup_requires,
    extras_require=extras_require,
    tests_require=[
        "pytest",
    ],
    project_urls={
        "Github": "https://github.com/oracle/accelerated-data-science",
        "Documentation": "https://accelerated-data-science.readthedocs.io/en/latest/index.html",
    },
    scripts=["ads/ads"],
)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
from __future__ import annotations
import re
from typing import Dict, TypeVar
from ads.jobs.builders.base import Builder
from ads.jobs import env_var_parser


Self = TypeVar("Self", bound="Runtime")
"""Special type to represent the current enclosed class.

This type is used by factory class method or when a method returns ``self``.
"""


class Runtime(Builder):
    """Base class for job runtime"""

    # Constant strings
    CONST_ENV_VAR = "env"
    CONST_ARGS = "args"
    CONST_MAXIMUM_RUNTIME_IN_MINUTES = "maximumRuntimeInMinutes"
    CONST_TAG = "freeformTags"

    attribute_map = {
        CONST_TAG: "freeform_tags",
        CONST_ENV_VAR: CONST_ENV_VAR,
    }

    def __init__(self, spec: Dict = None, **kwargs) -> None:
        env = {}
        if spec and "env" in spec and isinstance(spec["env"], dict):
            env = spec.pop("env", {})
        if "env" in kwargs:
            env = kwargs.pop("env")
        super().__init__(spec, **kwargs)
        if isinstance(env, dict):
            self.with_environment_variable(**env)

    @property
    def kind(self) -> str:
        """Kind of the object to be stored in YAML.
        All runtime implementations will have "runtime" as kind.
        Subclass will have different types.
        """
        return "runtime"

    @property
    def type(self) -> str:
        """The type of the object as showing in YAML"""
        class_name = self.__class__.__name__
        if class_name.endswith("Runtime") and class_name != "Runtime":
            class_name = re.sub(r"Runtime$", "", class_name)
        return class_name[0].lower() + (class_name[1:] if len(class_name) > 1 else "")

    def with_argument(self: Self, *args, **kwargs) -> Self:
        """Adds command line arguments to the runtime.

        This method can be called (chained) multiple times to add various arguments.

        Parameters
        ----------
        args:
            Positional arguments.
            In a single method call, positional arguments are always added before keyword arguments.
            You can call with_argument() to add positional arguments after keyword arguments.

        kwargs:
            Keyword arguments.
            To add a keyword argument without value, set the value to None.

        Returns
        -------
        Self
            This method returns self to support chaining methods.

        Raises
        ------
        ValueError
            Keyword arguments with space in a key.

        Examples
        --------
        >>> runtime = Runtime().with_argument(key1="val1", key2="val2").with_argument("pos1")
        >>> print(runtime.args)
        ["--key1", "val1", "--key2", "val2", "pos1"]

        >>> runtime = Runtime()
        >>> runtime.with_argument("pos1")
        >>> runtime.with_argument(key1="val1", key2="val2.1 val2.2")
        >>> runtime.with_argument("pos2")
        >>> print(runtime.args)
        ['pos1', '--key1', 'val1', '--key2', 'val2.1 val2.2', 'pos2']

        >>> runtime = Runtime()
        >>> runtime.with_argument("pos1")
        >>> runtime.with_argument(key1=None, key2="val2")
        >>> runtime.with_argument("pos2")
        >>> print(runtime.args)
        ["pos1", "--key1", "--key2", "val2", "pos2"]

        """
        arg_values = self.get_spec(self.CONST_ARGS, [])
        args = [str(arg) for arg in args]
        arg_values.extend(args)
        for k, v in kwargs.items():
            if " " in k:
                raise ValueError("Argument key cannot contain space.")
            arg_values.append(f"--{str(k)}")
            # Ignore None value
            if v is None:
                continue
            arg_values.append(str(v))
        self.set_spec(self.CONST_ARGS, arg_values)
        return self

    def with_environment_variable(self: Self, **kwargs) -> Self:
        """Sets environment variables

        Environment variables enclosed by ``${...}`` will be substituted.

        * You can use ``$$`` to escape the substitution.
        * Undefined variable enclosed by ``${}`` will be ignored.
        * Double dollar signs ``$$`` will be substituted by a single one ``$``.

        Returns
        -------
        Self
            This method returns self to support chaining methods.

        Examples
        --------
        >>> runtime = (
        ...     PythonRuntime()
        ...     .with_environment_variable(
        ...         HOST="10.0.0.1",
        ...         PORT="443",
        ...         URL="http://${HOST}:${PORT}/path/",
        ...         ESCAPED_URL="http://$${HOST}:$${PORT}/path/",
        ...         MISSING_VAR="This is ${UNDEFINED}",
        ...         VAR_WITH_DOLLAR="$10",
        ...         DOUBLE_DOLLAR="$$10"
        ...     )
        ... )
        >>> for k, v in runtime.environment_variables.items():
        ...     print(f"{k}: {v}")
        HOST: 10.0.0.1
        PORT: 443
        URL: http://10.0.0.1:443/path/
        ESCAPED_URL: http://${HOST}:${PORT}/path/
        MISSING_VAR: This is ${UNDEFINED}
        VAR_WITH_DOLLAR: $10
        DOUBLE_DOLLAR: $10


        """
        if not kwargs:
            return self
        envs = [{"name": k, "value": v} for k, v in kwargs.items()]
        return self.set_spec(self.CONST_ENV_VAR, envs)

    def with_freeform_tag(self: Self, **kwargs) -> Self:
        """Sets freeform tag

        Returns
        -------
        Self
            This method returns self to support chaining methods.
        """
        return self.set_spec(self.CONST_TAG, kwargs)

    def with_maximum_runtime_in_minutes(
        self: Self, maximum_runtime_in_minutes: int
    ) -> Self:
        """Sets maximum runtime in minutes

        Returns
        -------
        Self
            This method returns self to support chaining methods.
        """
        return self.set_spec(
            self.CONST_MAXIMUM_RUNTIME_IN_MINUTES, maximum_runtime_in_minutes
        )

    @property
    def environment_variables(self) -> dict:
        """Environment variables

        Returns
        -------
        dict
            The runtime environment variables.
            The returned dictionary is a copy.
        """
        env_var_list = self.get_spec(self.CONST_ENV_VAR)
        if env_var_list:
            return env_var_parser.parse(env_var_list)
        return {}

    @property
    def envs(self) -> dict:
        """Environment variables"""
        return self.environment_variables

    @property
    def freeform_tags(self) -> dict:
        """freeform_tags"""
        return self.get_spec(self.CONST_TAG, {})

    @property
    def args(self) -> list:
        """Command line arguments"""
        return self.get_spec(self.CONST_ARGS, [])

    @property
    def maximum_runtime_in_minutes(self) -> int:
        """Maximum runtime in minutes"""
        return self.get_spec(self.CONST_MAXIMUM_RUNTIME_IN_MINUTES)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


import click

import ads
from ads.common.auth import AuthType, AuthContext
from ads.jobs import Job, DataFlowRun, DataScienceJobRun
from ads.opctl.utils import suppress_traceback


@click.group("jobs")
@click.help_option("--help", "-h")
def commands():
    pass


@commands.command()
@click.help_option("--help", "-h")
@click.option("--file", "-f", help="path to YAML file", required=True)
@click.option(
    "--auth",
    "-a",
    help="authentication method",
    type=click.Choice(AuthType.values()),
    default="api_key",
)
@click.option(
    "--oci-profile", help="oci config profile", required=False, default="DEFAULT"
)
@click.option("--debug", "-d", help="set debug mode", is_flag=True, default=False)
def run(file, auth, oci_profile, debug):
    def _run(file, auth, oci_profile):
        with AuthContext():
            ads.set_auth(auth=auth, profile=oci_profile)
            job = Job.from_yaml(uri=file)
            job.create()
            print(job.id)
            job_run = job.run()
            print(job_run.id)

    suppress_traceback(debug)(_run)(file, auth, oci_profile)


@commands.command()
@click.help_option("--help", "-h")
@click.argument("ocid")
@click.option(
    "--auth",
    "-a",
    help="authentication method",
    type=click.Choice(AuthType.values()),
    default="api_key",
)
@click.option(
    "--oci-profile", help="oci config profile", required=False, default="DEFAULT"
)
@click.option("--debug", "-d", help="set debug mode", is_flag=True, default=False)
def watch(ocid, auth, oci_profile, debug):
    def _watch(ocid, auth, oci_profile):
        with AuthContext():
            ads.set_auth(auth=auth, profile=oci_profile)
            if "datasciencejobrun" in ocid:
                run = DataScienceJobRun.from_ocid(ocid)
            elif "dataflowrun" in ocid:
                run = DataFlowRun.from_ocid(ocid)
            else:
                raise ValueError("Run OCID not recognized.")
            run.watch()

    suppress_traceback(debug)(_watch)(ocid, auth, oci_profile)


@commands.command()
@click.help_option("--help", "-h")
@click.argument("ocid")
@click.option(
    "--auth",
    "-a",
    help="authentication method",
    type=click.Choice(AuthType.values()),
    default="api_key",
)
@click.option(
    "--oci-profile", help="oci config profile", required=False, default="DEFAULT"
)
@click.option("--debug", "-d", help="set debug mode", is_flag=True, default=False)
def delete(ocid, auth, oci_profile, debug):
    def _delete(ocid, auth, oci_profile):
        with AuthContext():
            ads.set_auth(auth=auth, profile=oci_profile)
            if "datasciencejobrun" in ocid:
                run = DataScienceJobRun.from_ocid(ocid)
            elif "dataflowrun" in ocid:
                run = DataFlowRun.from_ocid(ocid)
            else:
                raise ValueError("Run OCID not recognized.")
            run.delete()

    suppress_traceback(debug)(_delete)(ocid, auth, oci_profile)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import json
from typing import Any, Dict, Generator

import fsspec
from ads.data_labeling.interface.reader import Reader
from ads.common import auth as authutil


class JsonlReader(Reader):
    """JsonlReader class which reads the file."""

    def __init__(self, path: str, auth: Dict = None, encoding="utf-8") -> "JsonlReader":
        """Initiates a JsonlReader object.

        Parameters
        ----------
        path: str
            object storage path or local path for a file.
        auth: (dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.
        encoding : (str, optional). Defaults to 'utf-8'.
            Encoding of files. Only used for "TEXT" dataset.

        Examples
        --------
        >>> from ads.data_labeling.reader.jsonl_reader import JsonlReader
        >>> path = "your/path/to/jsonl/file.jsonl"
        >>> from ads.common import auth as authutil
        >>> reader = JsonlReader(path=path, auth=authutil.api_keys(), encoding="utf-8")
        >>> next(reader.read())
        """
        self.path = path
        self.auth = auth or authutil.default_signer()
        self.encoding = encoding

    def read(self, skip: int = None) -> Generator[Dict, Any, Any]:
        """Reads and yields the content of the file.

        Parameters
        ----------
        skip: (int, optional). Defaults to None.
            The number of records that should be skipped.

        Returns
        -------
        Generator[Dict, Any, Any]
            The content of the file.

        Raises
        ------
        ValueError
            If `skip` not empty and not a positive integer.
        FileNotFoundError
            When file not found.
        """
        if skip and (not isinstance(skip, int) or skip < 1):
            raise ValueError("The parameter `skip` must be a positive integer.")
        try:
            line_number = 0
            with fsspec.open(self.path, "r", encoding=self.encoding, **self.auth) as f:
                for line in f:
                    line_number += 1
                    if skip and line_number <= skip:
                        continue
                    yield json.loads(line)
        except FileNotFoundError:
            raise FileNotFoundError(f"Path ({self.path}) not found.")



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import logging

logger = logging.getLogger(__name__)

from ads.common import auth



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
try:
    import click
    import ads.opctl.cli
    import ads.jobs.cli
    import ads.pipeline.cli
    import os
    import json
except:
    print(
        "Please run `pip install oracle-ads[opctl]` to install the required dependencies for ADS CLI"
    )
    exit()


with open(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), "ads_version.json")
) as version_file:
    ADS_VERSION = json.load(version_file)["version"]


@click.group()
@click.version_option(version=ADS_VERSION, prog_name="ads")
@click.help_option("--help", "-h")
def cli():
    pass


cli.add_command(ads.opctl.cli.commands)
cli.add_command(ads.jobs.cli.commands)
cli.add_command(ads.pipeline.cli.commands)


if __name__ == "__main__":
    cli()



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import json
import logging
import os
import sys
from abc import ABC, abstractclassmethod, abstractmethod
from dataclasses import dataclass, field, fields
from pathlib import Path
from typing import Dict, List, Tuple

import ads.dataset.factory as factory
import fsspec
import git
import oci.data_science.models
import pandas as pd
import yaml
from ads.common import logger
from ads.common.error import ChangesNotCommitted
from ads.common.extended_enum import ExtendedEnumMeta
from ads.common.serializer import DataClassSerializable
from oci.util import to_dict

try:
    from yaml import CDumper as dumper
except:
    from yaml import Dumper as dumper


logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger("ADS")

METADATA_SIZE_LIMIT = 32000
METADATA_VALUE_LENGTH_LIMIT = 255
METADATA_DESCRIPTION_LENGTH_LIMIT = 255
_METADATA_EMPTY_VALUE = "NA"
CURRENT_WORKING_DIR = "."


class MetadataSizeTooLarge(ValueError):
    """Maximum allowed size for model metadata has been exceeded.
    See https://docs.oracle.com/en-us/iaas/data-science/using/models_saving_catalog.htm for more details.
    """

    def __init__(self, size: int):
        super().__init__(
            f"The metadata is `{size}` bytes and exceeds "
            f"the size limit of `{METADATA_SIZE_LIMIT}` bytes. "
            "Reduce the metadata size."
        )


class MetadataValueTooLong(ValueError):
    """Maximum allowed length of metadata value has been exceeded.
    See https://docs.oracle.com/en-us/iaas/data-science/using/models_saving_catalog.htm for more details.
    """

    def __init__(self, key: str, length: int):
        super().__init__(
            f"The custom metadata value of `{key}` is `{length}` characters and exceeds "
            f"the length limit of `{METADATA_VALUE_LENGTH_LIMIT}` characters."
        )


class MetadataDescriptionTooLong(ValueError):
    """Maximum allowed length of metadata description has been exceeded.
    See https://docs.oracle.com/en-us/iaas/data-science/using/models_saving_catalog.htm for more details.
    """

    def __init__(self, key: str, length: int):
        super().__init__(
            f"The custom metadata description of `{key}` is `{length}` characters and exceeds "
            f"the length limit of `{METADATA_DESCRIPTION_LENGTH_LIMIT}` characters."
        )


class MetadataCustomPrintColumns(str, metaclass=ExtendedEnumMeta):
    KEY = "Key"
    VALUE = "Value"
    DESCRIPTION = "Description"
    CATEGORY = "Category"


class MetadataTaxonomyPrintColumns(str, metaclass=ExtendedEnumMeta):
    KEY = "Key"
    VALUE = "Value"


class MetadataTaxonomyKeys(str, metaclass=ExtendedEnumMeta):
    USE_CASE_TYPE = "UseCaseType"
    FRAMEWORK = "Framework"
    FRAMEWORK_VERSION = "FrameworkVersion"
    ALGORITHM = "Algorithm"
    HYPERPARAMETERS = "Hyperparameters"
    ARTIFACT_TEST_RESULT = "ArtifactTestResults"


class MetadataCustomKeys(str, metaclass=ExtendedEnumMeta):
    SLUG_NAME = "SlugName"
    CONDA_ENVIRONMENT = "CondaEnvironment"
    CONDA_ENVIRONMENT_PATH = "CondaEnvironmentPath"
    ENVIRONMENT_TYPE = "EnvironmentType"
    MODEL_ARTIFACTS = "ModelArtifacts"
    TRAINING_DATASET = "TrainingDataset"
    VALIDATION_DATASET = "ValidationDataset"
    MODEL_SERIALIZATION_FORMAT = "ModelSerializationFormat"
    TRAINING_DATASET_SIZE = "TrainingDatasetSize"
    VALIDATION_DATASET_SIZE = "ValidationDatasetSize"
    TRAINING_DATASET_NUMBER_OF_ROWS = "TrainingDatasetNumberOfRows"
    TRAINING_DATASET_NUMBER_OF_COLS = "TrainingDatasetNumberOfCols"
    VALIDATION_DATASET_NUMBER_OF_ROWS = "ValidationDatasetNumberOfRows"
    VALIDATION_DATASET_NUMBER_OF_COLS = "ValidationDataSetNumberOfCols"
    CLIENT_LIBRARY = "ClientLibrary"
    MODEL_FILE_NAME = "ModelFileName"


class MetadataCustomCategory(str, metaclass=ExtendedEnumMeta):
    PERFORMANCE = "Performance"
    TRAINING_PROFILE = "Training Profile"
    TRAINING_AND_VALIDATION_DATASETS = "Training and Validation Datasets"
    TRAINING_ENV = "Training Environment"
    OTHER = "Other"


class UseCaseType(str, metaclass=ExtendedEnumMeta):
    BINARY_CLASSIFICATION = "binary_classification"
    REGRESSION = "regression"
    MULTINOMIAL_CLASSIFICATION = "multinomial_classification"
    CLUSTERING = "clustering"
    RECOMMENDER = "recommender"
    DIMENSIONALITY_REDUCTION = "dimensionality_reduction/representation"
    TIME_SERIES_FORECASTING = "time_series_forecasting"
    ANOMALY_DETECTION = "anomaly_detection"
    TOPIC_MODELING = "topic_modeling"
    NER = "ner"
    SENTIMENT_ANALYSIS = "sentiment_analysis"
    IMAGE_CLASSIFICATION = "image_classification"
    OBJECT_LOCALIZATION = "object_localization"
    OTHER = "other"


class Framework(str, metaclass=ExtendedEnumMeta):
    SCIKIT_LEARN = "scikit-learn"
    XGBOOST = "xgboost"
    TENSORFLOW = "tensorflow"
    PYTORCH = "pytorch"
    MXNET = "mxnet"
    KERAS = "keras"
    LIGHT_GBM = "lightgbm"
    PYMC3 = "pymc3"
    PYOD = "pyod"
    SPACY = "spacy"
    PROPHET = "prophet"
    SKTIME = "sktime"
    STATSMODELS = "statsmodels"
    CUML = "cuml"
    ORACLE_AUTOML = "oracle_automl"
    H20 = "h2o"
    TRANSFORMERS = "transformers"
    NLTK = "nltk"
    EMCEE = "emcee"
    PYSTAN = "pystan"
    BERT = "bert"
    GENSIM = "gensim"
    FLAIR = "flair"
    WORD2VEC = "word2vec"
    ENSEMBLE = "ensemble"
    SPARK = "pyspark"
    OTHER = "other"


class ModelMetadataItem(ABC):
    """The base abstract class representing model metadata item.

    Methods
    -------
    to_dict(self) -> Dict
        Serializes model metadata item to dictionary.
    from_dict(cls, data: Dict) -> ModelMetadataItem
        Constructs an instance of ModelMetadataItem from a dictionary.
    to_yaml(self)
        Serializes model metadata item to YAML.
    size(self) -> int
        Returns the size of the metadata in bytes.
    to_json(self) -> JSON
        Serializes metadata item to JSON.
    to_json_file(self, file_path: str, storage_options: dict = None) -> None
        Saves the metadata item value to a local file or object storage.
    validate(self) -> bool
        Validates metadata item.
    """

    _FIELDS = []

    @classmethod
    def from_dict(cls, data: Dict) -> "ModelMetadataItem":
        """Constructs an instance of `ModelMetadataItem` from a dictionary.

        Parameters
        ----------
        data : Dict
            Metadata item in a dictionary format.

        Returns
        -------
        ModelMetadataItem
            An instance of model metadata item.
        """
        return cls(**data or {})

    def to_dict(self) -> dict:
        """Serializes model metadata item to dictionary.

        Returns
        -------
        dict
            The dictionary representation of model metadata item.
        """
        return {field: getattr(self, field) for field in self._FIELDS}

    def to_yaml(self):
        """Serializes model metadata item to YAML.

        Returns
        -------
        Yaml
            The model metadata item in a YAML representation.
        """
        return yaml.dump(self.to_dict(), Dumper=dumper)

    def size(self) -> int:
        """Returns the size of the model metadata in bytes.

        Returns
        -------
        int
            The size of model metadata in bytes.
        """
        return len(json.dumps(self.to_dict()).encode("utf-16"))

    def to_json(self):
        """Serializes metadata item into a JSON.

        Returns
        -------
        JSON
            The metadata item in a JSON representation.
        """
        return json.dumps(self.to_dict())

    def to_json_file(
        self,
        file_path: str,
        storage_options: dict = None,
    ) -> None:
        """Saves the metadata item value to a local file or object storage.

        Parameters
        ----------
        file_path : str
            The file path to store the data.
            "oci://bucket_name@namespace/folder_name/"
            "oci://bucket_name@namespace/folder_name/result.json"
            "path/to/local/folder"
            "path/to/local/folder/result.json"
        storage_options : dict. Default None
            Parameters passed on to the backend filesystem class.
            Defaults to `options` set using `DatasetFactory.set_default_storage()`.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
            ValueError: When file path is empty.
            TypeError: When file path not a string.

        Examples
        --------
        >>> metadata_item = ModelCustomMetadataItem(key="key1", value="value1")
        >>> storage_options = {"config": oci.config.from_file(os.path.join("~/.oci", "config"))}
        >>> storage_options
        {'log_requests': False,
            'additional_user_agent': '',
            'pass_phrase': None,
            'user': '<user-id>',
            'fingerprint': '05:15:2b:b1:46:8a:32:ec:e2:69:5b:32:01:**:**:**)',
            'tenancy': '<tenency-id>',
            'region': 'us-ashburn-1',
            'key_file': '/home/datascience/.oci/oci_api_key.pem'}
        >>> metadata_item.to_json_file(file_path = 'oci://bucket_name@namespace/folder_name/file.json', storage_options=storage_options)
        >>> metadata_item.to_json_file("path/to/local/folder/file.json")
        """
        if not file_path:
            raise ValueError("File path must be specified.")

        if not isinstance(file_path, str):
            raise TypeError("File path must be a string.")

        if not Path(os.path.basename(file_path)).suffix:
            file_path = os.path.join(file_path, f"{self.key}.json")

        if not storage_options:
            storage_options = factory.default_storage_options or {"config": {}}

        with fsspec.open(
            file_path,
            mode="w",
            **(storage_options),
        ) as f:
            f.write(json.dumps(self.value))

    def _to_oci_metadata(self):
        """Converts metadata item to OCI metadata item."""
        dict = self.to_dict()
        if not dict["value"]:
            return oci.data_science.models.Metadata(**dict)
        if isinstance(dict["value"], (str, int, float)):
            dict["value"] = str(dict["value"]).replace("NaN", "null")
        else:
            dict["value"] = json.dumps(dict["value"]).replace("NaN", "null")
        return oci.data_science.models.Metadata(**dict)

    @classmethod
    def _from_oci_metadata(cls, oci_metadata_item) -> "ModelMetadataItem":
        """Creates a new metadata item from the OCI metadata item."""
        oci_metadata_item = to_dict(oci_metadata_item)
        key_value_map = {field: oci_metadata_item.get(field) for field in cls._FIELDS}

        if isinstance(key_value_map["value"], str):
            try:
                key_value_map["value"] = json.loads(oci_metadata_item.get("value"))
            except Exception:
                pass

        return cls(**key_value_map)

    def __hash__(self):
        return hash(self.key.lower())

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __repr__(self):
        return self.to_yaml()

    @abstractmethod
    def validate(self) -> bool:
        """Validates metadata item.

        Returns
        -------
        bool
            True if validation passed.
        """
        pass


class ModelTaxonomyMetadataItem(ModelMetadataItem):
    """Class that represents model taxonomy metadata item.

    Attributes
    ----------
    key: str
        The model metadata item key.
    value: str
        The model metadata item value.

    Methods
    -------
    reset(self) -> None
        Resets model metadata item.
    to_dict(self) -> Dict
        Serializes model metadata item to dictionary.
    from_dict(cls) -> ModelTaxonomyMetadataItem
        Constructs model metadata item from dictionary.
    to_yaml(self)
        Serializes model metadata item to YAML.
    size(self) -> int
        Returns the size of the metadata in bytes.
    update(self, value: str = "") -> None
        Updates metadata item information.
    to_json(self) -> JSON
        Serializes metadata item into a JSON.
    to_json_file(self, file_path: str, storage_options: dict = None) -> None
        Saves the metadata item value to a local file or object storage.
    validate(self) -> bool
        Validates metadata item.
    """

    _FIELDS = ["key", "value"]

    def __init__(
        self,
        key: str,
        value: str = None,
    ):
        self.key = key
        self.value = value

    @property
    def key(self) -> str:
        return self._key

    @key.setter
    def key(self, key: str):
        """The model metadata key setter.

        Raises
        ------
        TypeError
            If provided key is not a string.
        ValueError
            If provided key is already setup.
            If provided key is empty.
        """
        if hasattr(self, "_key"):
            raise ValueError("The key field is immutable and cannot be changed.")
        if not isinstance(key, str):
            raise TypeError("The key must be a string.")
        if key is None or key == "":
            raise ValueError("The key cannot be empty.")
        self._key = key

    @property
    def value(self) -> str:
        return self._value

    @value.setter
    def value(self, value: str):
        """The model metadata value setter. Accepts any JSON serializable value.

        Raises
        ------
        ValueError
            If provided value cannot be serialized to JSON.
        """
        if value is None or value == "":
            self._value = value
            return

        try:
            json.dumps(value)
        except TypeError:
            raise ValueError(
                f"An error occurred in attempt to serialize the value of {self.key} to JSON. "
                "The value must be JSON serializable."
            )

        self._value = value

    def reset(self) -> None:
        """Resets model metadata item.

        Resets value to None.

        Returns
        -------
        None
            Nothing.
        """
        self.update(value=None)

    def update(self, value: str) -> None:
        """Updates metadata item value.

        Parameters
        ----------
        value: str
            The value of model metadata item.

        Returns
        -------
        None
            Nothing.
        """
        self.value = value

    def validate(self) -> bool:
        """Validates metadata item.

        Returns
        -------
        bool
            True if validation passed.

        Raises
        ------
        ValueError
            If invalid UseCaseType provided.
            If invalid Framework provided.
        """
        if (
            self.key.lower() == MetadataTaxonomyKeys.USE_CASE_TYPE.lower()
            and self.value
            and (not isinstance(self.value, str) or self.value not in UseCaseType)
        ):
            raise ValueError(
                f"Invalid value of `UseCaseType`. Choose from {UseCaseType.values()}."
            )
        if (
            self.key.lower() == MetadataTaxonomyKeys.FRAMEWORK.lower()
            and self.value
            and (not isinstance(self.value, str) or self.value not in Framework)
        ):
            raise ValueError(
                f"Invalid value of `Framework`. Choose from {Framework.values()}."
            )
        return True


class ModelCustomMetadataItem(ModelTaxonomyMetadataItem):
    """Class that represents model custom metadata item.

    Attributes
    ----------
    key: str
        The model metadata item key.
    value: str
        The model metadata item value.
    description: str
        The model metadata item description.
    category: str
        The model metadata item category.

    Methods
    -------
    reset(self) -> None
        Resets model metadata item.
    to_dict(self)->dict
        Serializes model metadata item to dictionary.
    from_dict(cls) -> ModelCustomMetadataItem
        Constructs model metadata item from dictionary.
    to_yaml(self)
        Serializes model metadata item to YAML.
    size(self) -> int
        Returns the size of the metadata in bytes.
    update(self, value: str = "", description: str = "", category: str = "") -> None
        Updates metadata item information.
    to_json(self) -> JSON
        Serializes metadata item into a JSON.
    to_json_file(self, file_path: str, storage_options: dict = None) -> None
        Saves the metadata item value to a local file or object storage.
    validate(self) -> bool
        Validates metadata item.
    """

    _FIELDS = ["key", "value", "description", "category"]

    def __init__(
        self,
        key: str,
        value: str = None,
        description: str = None,
        category: str = None,
    ):
        super().__init__(key=key, value=value)
        self.description = description
        self.category = category

    @property
    def description(self) -> str:
        return self._description

    @description.setter
    def description(self, description: str):
        """The model metadata description setter.

        Raises
        ------
        TypeError
            If provided key is not a string.
        """
        if description != None and not isinstance(description, str):
            raise TypeError("The description must be a string.")

        self._description = description

    @property
    def category(self) -> str:
        return self._category

    @category.setter
    def category(self, category: str):
        """The model metadata category setter.

        Raises
        ------
        TypeError
            If provided category is not a string.
        ValueError
            If provided category not supported.
        """
        if not category:
            self._category = None
            return

        if not isinstance(category, str):
            raise TypeError(
                f"Invalid category type for the {self.key}."
                "The category must be a string."
            )

        if category not in MetadataCustomCategory:
            raise ValueError(
                f"Invalid category value for the {self.key}. "
                f"Choose from {MetadataCustomCategory.values()}."
            )

        self._category = category

    def reset(self) -> None:
        """Resets model metadata item.

        Resets value, description and category to None.

        Returns
        -------
        None
            Nothing.
        """
        self.update(value=None, description=None, category=None)

    def update(self, value: str, description: str, category: str) -> None:
        """Updates metadata item.

        Parameters
        ----------
        value: str
            The value of model metadata item.
        description: str
            The description of model metadata item.
        category: str
            The category of model metadata item.

        Returns
        -------
        None
            Nothing.
        """
        self.value = value
        self.description = description
        self.category = category

    def _to_oci_metadata(self):
        """Converts metadata item to OCI metadata item."""
        oci_metadata_item = super()._to_oci_metadata()
        if not oci_metadata_item.value:
            oci_metadata_item.value = _METADATA_EMPTY_VALUE
        if not oci_metadata_item.category:
            oci_metadata_item.category = MetadataCustomCategory.OTHER
        return oci_metadata_item

    def validate(self) -> bool:
        """Validates metadata item.

        Returns
        -------
        bool
            True if validation passed.

        Raises
        ------
        ValueError
            If invalid category provided.
        MetadataValueTooLong
            If value exceeds the length limit.
        """
        if self.category and self.category not in MetadataCustomCategory:
            raise ValueError(
                f"Invalid category value for the {self.key}. "
                f"Choose from {MetadataCustomCategory.values()}."
            )

        if self.value:
            value = (
                self.value if isinstance(self.value, str) else json.dumps(self.value)
            )
            if len(value) > METADATA_VALUE_LENGTH_LIMIT:
                raise MetadataValueTooLong(self.key, len(value))

        if (
            self.description
            and len(self.description) > METADATA_DESCRIPTION_LENGTH_LIMIT
        ):
            raise MetadataDescriptionTooLong(self.key, len(self.description))

        return True


class ModelMetadata(ABC):
    """The base abstract class representing model metadata.

    Methods
    -------
    get(self, key: str) -> ModelMetadataItem
        Returns the model metadata item by provided key.
    reset(self) -> None
        Resets all model metadata items to empty values.
    to_dataframe(self) -> pd.DataFrame
        Returns the model metadata list in a data frame format.
    size(self) -> int
        Returns the size of the model metadata in bytes.
    validate(self) -> bool
        Validates metadata.
    to_dict(self)
        Serializes model metadata into a dictionary.
    from_dict(cls) -> ModelMetadata
        Constructs model metadata from dictionary.
    to_yaml(self)
        Serializes model metadata into a YAML.
    to_json(self)
        Serializes model metadata into a JSON.
    to_json_file(self, file_path: str, storage_options: dict = None) -> None
        Saves the metadata to a local file or object storage.
    """

    @abstractmethod
    def __init__(self):
        """Initializes Model Metadata."""
        self._items = set()

    def get(self, key: str) -> ModelMetadataItem:
        """Returns the model metadata item by provided key.

        Parameters
        ----------
        key: str
            The key of model metadata item.

        Returns
        -------
        ModelMetadataItem
            The model metadata item.

        Raises
        ------
        ValueError
            If provided key is empty or metadata item not found.
        """
        if key is None or not isinstance(key, str) or key == "":
            raise ValueError("The key must not be an empty string.")
        for item in self._items:
            if item.key.lower() == key.lower():
                return item
        raise ValueError(f"The metadata with {key} not found.")

    def reset(self) -> None:
        """Resets all model metadata items to empty values.

        Resets value, description and category to None for every metadata item.
        """
        for item in self._items:
            item.reset()

    def size(self) -> int:
        """Returns the size of the model metadata in bytes.

        Returns
        -------
        int
            The size of model metadata in bytes.
        """
        return sum(item.size() for item in self._items)

    def validate_size(self) -> bool:
        """Validates model metadata size.

        Validates the size of metadata. Throws an error if the size of the metadata
        exceeds expected value.

        Returns
        -------
        bool
            True if metadata size is valid.

        Raises
        ------
        MetadataSizeTooLarge
            If the size of the metadata exceeds expected value.
        """
        if self.size() > METADATA_SIZE_LIMIT:
            raise MetadataSizeTooLarge(self.size())
        return True

    def validate(self) -> bool:
        """Validates model metadata.

        Returns
        -------
        bool
            True if metadata is valid.
        """
        for item in self._items:
            item.validate()
        return True

    def to_dict(self):
        """Serializes model metadata into a dictionary.

        Returns
        -------
        Dict
            The model metadata in a dictionary representation.
        """
        return {"data": [item.to_dict() for item in self._items]}

    def to_yaml(self):
        """Serializes model metadata into a YAML.

        Returns
        -------
        Yaml
            The model metadata in a YAML representation.
        """
        return yaml.dump(self.to_dict(), Dumper=dumper)

    def to_json(self):
        """Serializes model metadata into a JSON.

        Returns
        -------
        JSON
            The model metadata in a JSON representation.
        """
        return json.dumps(self.to_dict())

    @property
    def keys(self) -> Tuple[str]:
        """Returns all registered metadata keys.

        Returns
        -------
        Tuple[str]
            The list of metadata keys.
        """
        return tuple(item.key for item in self._items)

    def _to_oci_metadata(self):
        """Convert to a list of `oci.data_science.models.Metadata` objects.

        Returns
        -------
        list[oci.data_science.models.Metadata]
            A list of oci data science model metadata.

        Examples
        --------
        >>> metadata_taxonomy = ModelTaxonomyMetadata()
        >>> metadata_taxonomy.get(key="FrameworkVersion").update(value="2.3.1")
        >>> metadata_taxonomy._to_oci_metadata()
        [{
        "key": "FrameworkVersion",
        "value": "2.3.1"
        },
        {
        "key": "UseCaseType",
        "value": null
        },
        {
        "key": "Algorithm",
        "value": null
        },
        {
        "key": "Framework",
        "value": null
        },
        {
        "key": "Hyperparameters",
        "value": null
        }]
        """
        return [item._to_oci_metadata() for item in self._items]

    def to_json_file(
        self,
        file_path: str,
        storage_options: dict = None,
    ) -> None:
        """Saves the metadata to a local file or object storage.

        Parameters
        ----------
        file_path : str
            The file path to store the data.
            "oci://bucket_name@namespace/folder_name/"
            "oci://bucket_name@namespace/folder_name/metadata.json"
            "path/to/local/folder"
            "path/to/local/folder/metadata.json"
        storage_options : dict. Default None
            Parameters passed on to the backend filesystem class.
            Defaults to `options` set using `DatasetFactory.set_default_storage()`.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
            ValueError: When file path is empty.
            TypeError: When file path not a string.

        Examples
        --------
        >>> metadata = ModelTaxonomyMetadataItem()
        >>> storage_options = {"config": oci.config.from_file(os.path.join("~/.oci", "config"))}
        >>> storage_options
        {'log_requests': False,
            'additional_user_agent': '',
            'pass_phrase': None,
            'user': '<user-id>',
            'fingerprint': '05:15:2b:b1:46:8a:32:ec:e2:69:5b:32:01:**:**:**)',
            'tenancy': '<tenancy-id>',
            'region': 'us-ashburn-1',
            'key_file': '/home/datascience/.oci/oci_api_key.pem'}
        >>> metadata.to_json_file(file_path = 'oci://bucket_name@namespace/folder_name/metadata_taxonomy.json', storage_options=storage_options)
        >>> metadata_item.to_json_file("path/to/local/folder/metadata_taxonomy.json")
        """
        if not file_path:
            raise ValueError("File path must be specified.")

        if not isinstance(file_path, str):
            raise TypeError("File path must be a string.")

        if not Path(os.path.basename(file_path)).suffix:
            file_path = os.path.join(file_path, f"{self.__class__.__name__}.json")

        if not storage_options:
            storage_options = factory.default_storage_options or {"config": {}}

        with fsspec.open(
            file_path,
            mode="w",
            **(storage_options),
        ) as f:
            f.write(self.to_json())

    def __getitem__(self, key: str) -> ModelMetadataItem:
        return self.get(key)

    def __repr__(self):
        return self.to_yaml()

    def __len__(self):
        return len(self._items)

    @abstractclassmethod
    def _from_oci_metadata(cls, metadata_list):
        pass

    @abstractmethod
    def to_dataframe(self) -> pd.DataFrame:
        """Returns the model metadata list in a data frame format.

        Returns
        -------
        `pandas.DataFrame`
            The model metadata in a dataframe format.
        """
        pass

    @abstractclassmethod
    def from_dict(cls, data: Dict) -> "ModelMetadata":
        """Constructs an instance of `ModelMetadata` from a dictionary.

        Parameters
        ----------
        data : Dict
            Model metadata in a dictionary format.

        Returns
        -------
        ModelMetadata
            An instance of model metadata.
        """
        pass


class ModelCustomMetadata(ModelMetadata):
    """Class that represents Model Custom Metadata.

    Methods
    -------
    get(self, key: str) -> ModelCustomMetadataItem
        Returns the model metadata item by provided key.
    reset(self) -> None
        Resets all model metadata items to empty values.
    to_dataframe(self) -> pd.DataFrame
        Returns the model metadata list in a data frame format.
    size(self) -> int
        Returns the size of the model metadata in bytes.
    validate(self) -> bool
        Validates metadata.
    to_dict(self)
        Serializes model metadata into a dictionary.
    from_dict(cls) -> ModelCustomMetadata
        Constructs model metadata from dictionary.
    to_yaml(self)
        Serializes model metadata into a YAML.
    add(self, key: str, value: str, description: str = "", category: str = MetadataCustomCategory.OTHER, replace: bool = False) -> None:
        Adds a new model metadata item. Replaces existing one if replace flag is True.
    remove(self, key: str) -> None
        Removes a model metadata item by key.
    clear(self) -> None
        Removes all metadata items.
    isempty(self) -> bool
        Checks if metadata is empty.
    to_json(self)
        Serializes model metadata into a JSON.
    to_json_file(self, file_path: str, storage_options: dict = None) -> None
        Saves the metadata to a local file or object storage.

    Examples
    --------
    >>> metadata_custom = ModelCustomMetadata()
    >>> metadata_custom.add(key="format", value="pickle")
    >>> metadata_custom.add(key="note", value="important note", description="some description")
    >>> metadata_custom["format"].description = "some description"
    >>> metadata_custom.to_dataframe()
                        Key              Value         Description      Category
    ----------------------------------------------------------------------------
    0                format             pickle    some description  user defined
    1                  note     important note    some description  user defined
    >>> metadata_custom
        metadata:
        - category: user defined
          description: some description
          key: format
          value: pickle
        - category: user defined
          description: some description
          key: note
          value: important note
    >>> metadata_custom.remove("format")
    >>> metadata_custom
        metadata:
        - category: user defined
          description: some description
          key: note
          value: important note
    >>> metadata_custom.to_dict()
        {'metadata': [{
                'key': 'note',
                'value': 'important note',
                'category': 'user defined',
                'description': 'some description'
            }]}
    >>> metadata_custom.reset()
    >>> metadata_custom
        metadata:
        - category: None
          description: None
          key: note
          value: None
    >>> metadata_custom.clear()
    >>> metadata_custom.to_dataframe()
                        Key              Value         Description      Category
    ----------------------------------------------------------------------------
    """

    def __init__(self):
        """Initializes custom model metadata."""
        self._items = set()

    def add(
        self,
        key: str,
        value: str,
        description: str = "",
        category: str = MetadataCustomCategory.OTHER,
        replace: bool = False,
    ) -> None:
        """Adds a new model metadata item. Overrides the existing one if replace flag is True.

        Parameters
        ----------
        key: str
            The metadata item key.
        value: str
            The metadata item value.
        description: str
            The metadata item description.
        category: str
            The metadata item category.
        replace: bool
            Overrides the existing metadata item if replace flag is True.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
        TypeError
            If provided key is not a string.
            If provided description not a string.
        ValueError
            If provided key is empty.
            If provided value is empty.
            If provided value cannot be serialized to JSON.
            If item with provided key is already registered and replace flag is False.
            If provided category is not supported.
        MetadataValueTooLong
            If the length of provided value exceeds 255 charracters.
        MetadataDescriptionTooLong
            If the length of provided description exceeds 255 charracters.
        """

        if not category:
            category = MetadataCustomCategory.OTHER

        if key is None or key == "":
            raise ValueError("The key cannot be empty.")

        if value is None or value == "":
            raise ValueError("The value cannot be empty.")

        if not isinstance(key, str):
            raise TypeError("The key must be a string.")

        if not isinstance(category, str):
            raise TypeError("The category must be a string.")

        if category not in MetadataCustomCategory:
            raise ValueError(
                f"Invalid category value. "
                f"Choose from {MetadataCustomCategory.values()}."
            )

        if description and not isinstance(description, str):
            raise TypeError("The description must be a string.")

        try:
            tmp_value = json.dumps(value)
        except TypeError:
            raise ValueError(
                f"An error occurred in attempt to serialize the value of `{key}` to JSON. "
                "The value must be JSON serializable."
            )

        if len(tmp_value) > METADATA_VALUE_LENGTH_LIMIT:
            raise MetadataValueTooLong(key, len(tmp_value))

        if description and len(description) > METADATA_DESCRIPTION_LENGTH_LIMIT:
            raise MetadataDescriptionTooLong(key, len(description))

        if not replace and key in self.keys:
            raise ValueError(
                f"The metadata item with key {key} is already registered. "
                "Use replace=True to overwrite."
            )

        self._add(
            ModelCustomMetadataItem(
                key=key,
                value=value,
                description=description,
                category=category,
            ),
            replace=replace,
        )

    def _add(self, item: ModelCustomMetadataItem, replace=False) -> None:
        """Adds a new model metadata item.

        Overrides the existing one if replace flag is True.

        Parameters
        ----------
        item: ModelCustomMetadataItem
            The model metadata item.
        replace: bool
            Overrides the existing metadata item if replace flag is True.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
        ValueError
            If item is already registered and replace flag is False.
        TypeError
            If input data has a wrong format.
        """
        if not isinstance(item, ModelCustomMetadataItem):
            raise TypeError(
                "Argument must be an instance of the class ModelCustomMetadataItem."
            )
        if not replace and item in self._items:
            raise ValueError(
                f"The metadata item with key {item.key} is already registered. "
                "Use replace=True to overwrite."
            )
        self._items.discard(item)
        self._items.add(item)

    def _add_many(self, items: List[ModelCustomMetadataItem], replace=False) -> None:
        """Adds model metadata items into model metadata.

        Overrides the existing ones if replace flag is True.

        Parameters
        ----------
        items: List[ModelCustomMetadataItem]
            The list of model metadata items.
        replace: bool
            Overrides the existing metadata items if replace flag is True.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
        TypeError
            If input data has wrong format.
        """
        if not isinstance(items, list):
            raise TypeError("Argument must be a list of model metadata items.")
        for item in items:
            self._add(item, replace)

    def set_training_data(self, path: str, data_size: str = None):
        """Adds training_data path and data size information into model custom metadata.

        Parameters
        ----------
        path: str
            The path where the training_data is stored.
        data_size: str
            The size of the training_data.

        Returns
        -------
        None
            Nothing.
        """
        self.add(
            key=MetadataCustomKeys.TRAINING_DATASET,
            value=path,
            category=MetadataCustomCategory.TRAINING_AND_VALIDATION_DATASETS,
            description="The path where training dataset path are stored.",
            replace=True,
        )
        if data_size is not None:
            self.add(
                key=MetadataCustomKeys.TRAINING_DATASET_SIZE,
                value=data_size,
                category=MetadataCustomCategory.TRAINING_AND_VALIDATION_DATASETS,
                description="The size of the training data.",
                replace=True,
            )

    def set_validation_data(self, path: str, data_size: str = None):
        """Adds validation_data path and data size information into model custom metadata.

        Parameters
        ----------
        path: str
            The path where the validation_data is stored.
        data_size: str
            The size of the validation_data.

        Returns
        -------
        None
            Nothing.
        """
        self.add(
            key=MetadataCustomKeys.VALIDATION_DATASET,
            value=path,
            category=MetadataCustomCategory.TRAINING_AND_VALIDATION_DATASETS,
            description="The path where validation dataset path are stored.",
            replace=True,
        )
        if data_size is not None:
            self.add(
                key=MetadataCustomKeys.VALIDATION_DATASET_SIZE,
                value=data_size,
                category=MetadataCustomCategory.TRAINING_AND_VALIDATION_DATASETS,
                description="The size of the validation data.",
                replace=True,
            )

    def remove(self, key: str) -> None:
        """Removes a model metadata item.

        Parameters
        ----------
        key: str
            The key of the metadata item that should be removed.

        Returns
        -------
        None
            Nothing.
        """
        self._items.discard(self.get(key))

    def clear(self) -> None:
        """Removes all metadata items.

        Returns
        -------
        None
            Nothing.
        """
        self._items.clear()

    def isempty(self) -> bool:
        """Checks if metadata is empty.

        Returns
        -------
        bool
            True if metadata is empty, False otherwise.
        """
        return len(self._items) == 0

    @classmethod
    def _from_oci_metadata(cls, metadata_list):
        """Convert from list of OCI metadata list to an ModelCustomMetadata object.

        Returns
        -------
        ModelCustomMetadata
            A `ModelCustomMetadata` instance.
        """
        metadata = cls()
        for item in metadata_list:
            metadata._add(
                ModelCustomMetadataItem._from_oci_metadata(item), replace=True
            )
        return metadata

    def to_dataframe(self) -> pd.DataFrame:
        """Returns the model metadata list in a data frame format.

        Returns
        -------
        `pandas.DataFrame`
            The model metadata in a dataframe format.
        """
        return (
            pd.DataFrame(
                (
                    (item.key, item.value, item.description, item.category)
                    for item in self._items
                ),
                columns=[value for value in MetadataCustomPrintColumns.values()],
            )
            .sort_values(by=[MetadataCustomPrintColumns.KEY])
            .reset_index(drop=True)
        )

    @classmethod
    def from_dict(cls, data: Dict) -> "ModelCustomMetadata":
        """Constructs an instance of `ModelCustomMetadata` from a dictionary.

        Parameters
        ----------
        data : Dict
            Model metadata in a dictionary format.

        Returns
        -------
        ModelCustomMetadata
            An instance of model custom metadata.

        Raises
        ------
        ValueError
            In case of the wrong input data format.
        """
        if (
            not data
            or not isinstance(data, Dict)
            or not "data" in data
            or not isinstance(data["data"], List)
        ):
            raise ValueError(
                "An error occurred when attempting to deserialize the model custom metadata from a dictionary. "
                "The input data must be a dictionary with `data` key. Example: `{'data': []}`"
            )

        metadata = cls()
        for item in data["data"]:
            metadata._add(ModelCustomMetadataItem.from_dict(item), replace=True)
        return metadata


class ModelTaxonomyMetadata(ModelMetadata):
    """Class that represents Model Taxonomy Metadata.

    Methods
    -------
    get(self, key: str) -> ModelTaxonomyMetadataItem
        Returns the model metadata item by provided key.
    reset(self) -> None
        Resets all model metadata items to empty values.
    to_dataframe(self) -> pd.DataFrame
        Returns the model metadata list in a data frame format.
    size(self) -> int
        Returns the size of the model metadata in bytes.
    validate(self) -> bool
        Validates metadata.
    to_dict(self)
        Serializes model metadata into a dictionary.
    from_dict(cls) -> ModelTaxonomyMetadata
        Constructs model metadata from dictionary.
    to_yaml(self)
        Serializes model metadata into a YAML.
    to_json(self)
        Serializes model metadata into a JSON.
    to_json_file(self, file_path: str, storage_options: dict = None) -> None
        Saves the metadata to a local file or object storage.

    Examples
    --------
    >>> metadata_taxonomy = ModelTaxonomyMetadata()
    >>> metadata_taxonomy.to_dataframe()
                    Key                   Value
    --------------------------------------------
    0        UseCaseType   binary_classification
    1          Framework                 sklearn
    2   FrameworkVersion                   0.2.2
    3          Algorithm               algorithm
    4    Hyperparameters                      {}

    >>> metadata_taxonomy.reset()
    >>> metadata_taxonomy.to_dataframe()
                    Key                    Value
    --------------------------------------------
    0        UseCaseType                    None
    1          Framework                    None
    2   FrameworkVersion                    None
    3          Algorithm                    None
    4    Hyperparameters                    None

    >>> metadata_taxonomy
        metadata:
        - key: UseCaseType
          category: None
          description: None
          value: None
    """

    def __init__(self):
        super().__init__()
        for key in MetadataTaxonomyKeys.values():
            self._items.add(ModelTaxonomyMetadataItem(key=key))

    def _populate_from_map(self, map: Dict[str, str]) -> None:
        """Populates metadata information from map.

        Parameters
        ----------
        map: Dict[str, str]
            The key/value map with model metadata information.

        Returns
        -------
        None
            Nothing.
        """
        for value in MetadataTaxonomyKeys.values():
            if value in map:
                self[value].update(value=map[value])

    @classmethod
    def _from_oci_metadata(cls, metadata_list):
        """
        Convert from list of oci metadata to a ModelTaxonomyMetadata object.

        Parameters
        ----------
        metadata_list: List
            List of oci metadata.

        Returns
        -------
        ModelTaxonomyMetadata
            A `ModelTaxonomyMetadata` instance.
        """
        metadata = cls()
        for oci_item in metadata_list:
            item = ModelTaxonomyMetadataItem._from_oci_metadata(oci_item)
            metadata[item.key].update(value=item.value)
        return metadata

    def to_dataframe(self) -> pd.DataFrame:
        """Returns the model metadata list in a data frame format.

        Returns
        -------
        `pandas.DataFrame`
            The model metadata in a dataframe format.
        """
        return (
            pd.DataFrame(
                ((item.key, item.value) for item in self._items),
                columns=[value for value in MetadataTaxonomyPrintColumns.values()],
            )
            .sort_values(by=[MetadataTaxonomyPrintColumns.KEY])
            .reset_index(drop=True)
        )

    @classmethod
    def from_dict(cls, data: Dict) -> "ModelTaxonomyMetadata":
        """Constructs an instance of `ModelTaxonomyMetadata` from a dictionary.

        Parameters
        ----------
        data : Dict
            Model metadata in a dictionary format.

        Returns
        -------
        ModelTaxonomyMetadata
            An instance of model taxonomy metadata.

        Raises
        ------
        ValueError
            In case of the wrong input data format.
        """
        if (
            not data
            or not isinstance(data, Dict)
            or not "data" in data
            or not isinstance(data["data"], List)
        ):
            raise ValueError(
                "An error occurred when attempting to deserialize the model taxonomy metadata from a dictionary. "
                "The input data must be a dictionary with `data` key. Example: `{'data': []}`"
            )

        metadata = cls()
        for item in data["data"]:
            item = ModelTaxonomyMetadataItem.from_dict(item)
            metadata[item.key].update(value=item.value)
        return metadata


@dataclass(repr=True)
class ModelProvenanceMetadata(DataClassSerializable):
    """ModelProvenanceMetadata class.

    Examples
    --------
    >>> provenance_metadata = ModelProvenanceMetadata.fetch_training_code_details()
    ModelProvenanceMetadata(repo=<git.repo.base.Repo '/home/datascience/.git'>, git_branch='master', git_commit='99ad04c31803f1d4ffcc3bf4afbd6bcf69a06af2', repository_url='file:///home/datascience', "", "")
    >>> provenance_metadata.assert_path_not_dirty("your_path", ignore=False)
    """

    repo: str = field(default=None, metadata={"serializable": False})
    git_branch: str = None
    git_commit: str = None
    repository_url: str = None
    training_script_path: str = None
    training_id: str = None
    artifact_dir: str = None

    @classmethod
    def fetch_training_code_details(
        cls,
        training_script_path: str = None,
        training_id: str = None,
        artifact_dir: str = None,
    ):
        """Fetches the training code details: repo, git_branch, git_commit, repository_url, training_script_path and training_id.

        Parameters
        ----------
        training_script_path: (str, optional). Defaults to None.
            Training script path.
        training_id: (str, optional). Defaults to None.
            The training OCID for model.
        artifact_dir: str
            artifact directory to store the files needed for deployment.

        Returns
        -------
        ModelProvenanceMetadata
            A ModelProvenanceMetadata instance.
        """
        git_dir = CURRENT_WORKING_DIR
        if training_script_path:
            if not os.path.exists(training_script_path):
                logger.warning(
                    f"Training script {os.path.abspath(training_script_path)} does not exist."
                )
            else:
                training_script_path = os.path.abspath(training_script_path)
                git_dir = os.path.dirname(training_script_path)
        repo = git.Repo(git_dir, search_parent_directories=True)
        # get repository url
        if len(repo.remotes) > 0:
            repository_url = (
                repo.remotes.origin.url
                if repo.remotes.origin in repo.remotes
                else list(repo.remotes.values())[0].url
            )
        else:
            repository_url = "file://" + repo.working_dir  # no remote repo

        # get git commit
        git_branch = None
        git_commit = None

        try:
            # get git branch
            git_branch = format(repo.active_branch)
            git_commit = format(str(repo.head.commit.hexsha)) or None
        except Exception:
            logger.warning("No commit found.")
        return cls(
            repo=repo,
            git_branch=git_branch,
            git_commit=git_commit,
            repository_url=repository_url,
            training_script_path=training_script_path,
            training_id=training_id,
            artifact_dir=artifact_dir,
        )

    def assert_path_not_dirty(self, path: str, ignore: bool):
        """Checks if all the changes in this path has been commited.

        Parameters
        ----------
        path: (str)
            path.
        ignore (bool)
            whether to ignore the changes or not.

        Raises
        ------
        ChangesNotCommitted: if there are changes not being commited.

        Returns
        -------
        None
            Nothing.
        """
        if self.repo is not None and not ignore:
            path_abs = os.path.abspath(path)
            if (
                os.path.commonpath([path_abs, self.repo.working_dir])
                == self.repo.working_dir
            ):
                path_relpath = os.path.relpath(path_abs, self.repo.working_dir)
                if self.repo.is_dirty(path=path_relpath) or any(
                    [
                        os.path.commonpath([path_relpath, untracked]) == path_relpath
                        for untracked in self.repo.untracked_files
                    ]
                ):
                    raise ChangesNotCommitted(path_abs)

    def _to_oci_metadata(self) -> oci.data_science.models.ModelProvenance:
        """Convert to `oci.data_science.models.ModelProvenance` object.

        Returns
        -------
        oci.data_science.models.ModelProvenance
            OCI model provenance object.
        """
        return oci.data_science.models.ModelProvenance(
            repository_url=self.repository_url,
            git_branch=self.git_branch,
            git_commit=self.git_commit,
            script_dir=self.artifact_dir,
            training_script=self.training_script_path,
            training_id=self.training_id,
        )

    @classmethod
    def _from_oci_metadata(
        cls, model_provenance: oci.data_science.models.ModelProvenance
    ) -> "ModelProvenanceMetadata":
        """Creates a new model provenance metadata item from the `oci.data_science.models.ModelProvenance` object.

        Returns
        -------
        ModelProvenanceMetadata
            Model provenance metadata object.
        """
        return ModelProvenanceMetadata(
            repo=model_provenance.repository_url,
            git_branch=model_provenance.git_branch,
            git_commit=model_provenance.git_commit,
            repository_url=model_provenance.repository_url,
            training_script_path=model_provenance.training_script,
            training_id=model_provenance.training_id,
            artifact_dir=model_provenance.script_dir,
        )

    @classmethod
    def from_dict(cls, data: Dict[str, str]) -> "ModelProvenanceMetadata":
        """Constructs an instance of ModelProvenanceMetadata from a dictionary.

        Parameters
        ----------
        data : Dict[str,str]
            Model provenance metadata in dictionary format.

        Returns
        -------
        ModelProvenanceMetadata
            An instance of ModelProvenanceMetadata.
        """
        return cls(**data or {})

    def to_dict(self) -> dict:
        """Serializes model provenance metadata into a dictionary.

        Returns
        -------
        Dict
            The dictionary representation of the model provenance metadata.
        """
        return {
            f.name: getattr(self, f.name)
            for f in fields(self)
            if ("serializable" not in f.metadata or f.metadata["serializable"])
        }

    def __repr__(self):
        """Returns printable version of object.

        Parameters
        ----------
        string
            Serialized version of object as a YAML string
        """
        return self.to_yaml()



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from typing import Dict, List
from matplotlib.colors import LinearSegmentedColormap


def _sienna_light_to_dark_color_palette():
    return LinearSegmentedColormap.from_list("", ["#FDF3E4", "#512C1B"])


def _continuous_columns(feature_types: Dict) -> List[str]:
    """
    Parameters
    ----------
    feature_types : Dict
        Column name mapping to list of feature types ordered by most to least relevant.

    Returns
    -------
    List[str]
        List of columns that have continuous or ordinal in the feature type list.

    Note
    ____
    if a column has both ordinal/continuous and categorical pick whichever comes first.
    """
    continuous_cols = []
    for col in feature_types:
        for feature_type in feature_types[col]:
            if feature_type == "continuous" or feature_type == "ordinal":
                continuous_cols.append(col)
                break
            if feature_type == "category":
                break
    return continuous_cols


def _categorical_columns(feature_types: Dict) -> List[str]:
    """
    Parameters
    ----------
    feature_types : Dict
        column name mapping to list of feature types ordered by most to least relevant.

    Returns
    -------
    List[str]
        List of columns that have categorical in the feature type list.
    Note
    ____
    if a column has both ordinal/continuous and categorical pick whichever comes first.
    """
    categorical_cols = []
    for col in feature_types:
        for feature_type in feature_types[col]:
            if feature_type == "category":
                categorical_cols.append(col)
                break
            if feature_type == "continuous" or feature_type == "ordinal":
                break
    return categorical_cols



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
from typing import Any, Dict, Optional, Union
import warnings

import oci.data_science.models as data_science_models
from ads.common import utils
from ads.common.oci_datascience import OCIDataScienceMixin

from .common.utils import OCIClientManager


class ModelDeploymentProperties(
    OCIDataScienceMixin, data_science_models.ModelDeployment
):
    """Represents the details for a model deployment

    Attributes
    ----------
    swagger_types : dict
        The property names and the corresponding types of OCI ModelDeployment model.
    model_id : str
        The model artifact OCID in model catalog.
    model_uri : str
        uri to model files, can be local or in cloud storage.

    Methods
    -------
    with_prop(property_name, value)
        Set the model deployment details `property_name` attribute to `value`
    with_instance_configuration(config)
        Set the configuration of VM instance.
    with_access_log(log_group_id, log_id)
        Config the access log with OCI logging service
    with_predict_log(log_group_id, log_id)
        Config the predict log with OCI logging service
    build()
        Return an instance of CreateModelDeploymentDetails for creating the deployment.
    """

    # These properties are supported by ModelDeploymentProperties but are not top-level attributes of ModelDeployment
    sub_properties = [
        "instance_shape",
        "instance_count",
        "bandwidth_mbps",
        "access_log_group_id",
        "access_log_id",
        "predict_log_group_id",
        "predict_log_id",
        "memory_in_gbs",
        "ocpus",
    ]

    def __init__(
        self,
        model_id: Optional[str] = None,
        model_uri: Optional[str] = None,
        oci_model_deployment: Union[
            data_science_models.ModelDeployment,
            data_science_models.CreateModelDeploymentDetails,
            data_science_models.UpdateModelDeploymentDetails,
            Dict,
        ] = None,
        config: dict = None,
        **kwargs,
    ):
        """Initialize a ModelDeploymentProperties object by specifying one of the followings:

        Parameters
        ----------
        model_id: (str, optiona). Defaults to None.
            Model Artifact OCID. The model_id must be specified either
            explicitly or as an attribute of the OCI object.
        model_uri: (str, optiona). Defaults to None.
            Uri to model files, can be local or in cloud storage.
        oci_model_deployment: (Union[ModelDeployment, CreateModelDeploymentDetails, UpdateModelDeploymentDetails, Dict], optional). Defaults to None.
            An OCI model or Dict containing model deployment details.
            The OCI model can be an instance of either `ModelDeployment`,
            `CreateModelDeploymentDetails` or `UpdateModelConfigurationDetails`.
        config : (Dict, optional). Defaults to None.
            ADS auth dictionary for OCI authentication.
            This can be generated by calling ads.common.auth.api_keys() or
            ads.common.auth.resource_principal().
            If this is None, ads.common.default_signer(client_kwargs) will be used.
        kwargs:
            Users can also initialize the object by using keyword arguments.
            The following keyword arguments are supported by `oci.data_science.models.data_science_models.ModelDeployment`:

            - `display_name`,
            - `description`,
            - `project_id`,
            - `compartment_id`,
            - `model_deployment_configuration_details`,
            - `category_log_details`,
            - `freeform_tags`,
            - `defined_tags`.

            If display_name is not specified, a randomly generated easy to remember name will be generated,
            like 'strange-spider-2022-08-17-23:55.02'.

            ModelDeploymentProperties also supports the following additional
            keyward arguments:

            - `instance_shape`,
            - `instance_count`,
            - `bandwidth_mbps`,
            - `access_log_group_id`,
            - `access_log_id`,
            - `predict_log_group_id`,
            - `predict_log_id`,
            - `memory_in_gbs`,
            - `ocpus`.

            These additional arguments will be saved into appropriate properties
            in the OCI model.

        Raises
        ------
        ValueError
            model_id is None AND not specified in
            oci_model_deployment.model_deployment_configuration_details.model_configuration_details.
        """
        if config:
            warnings.warn(
                "`config` will be deprecated in 3.0.0 and will be ignored now."
            )

        self.model_id = model_id
        self.model_uri = model_uri

        oci_kwargs = {}
        self.oci_model_deployment = oci_model_deployment
        if oci_model_deployment:
            # User specified oci_model_deployment, which could be an OCI model object or a dict
            if isinstance(oci_model_deployment, dict):
                oci_kwargs = oci_model_deployment
                # If user specified properties defined in the sub_properties in oci_model_deployment (as dict),
                # move them to the kwargs so that they can be processed correctly.
                for key in list(oci_kwargs.keys()):
                    if key in self.sub_properties and not key in kwargs:
                        kwargs[key] = oci_kwargs.pop(key)
            else:
                # Extract properties from OCI model as a dict
                if hasattr(oci_model_deployment, "swagger_types"):
                    oci_kwargs = {
                        k: getattr(oci_model_deployment, k)
                        for k in oci_model_deployment.swagger_types.keys()
                    }
                elif oci_model_deployment:
                    raise ValueError(
                        "Invalid value for oci_model_deployment when initializing ModelDeploymentProperties: "
                        f"{oci_model_deployment}. "
                        "ModelDeploymentProperties should be initialized with an instance of OCI data science model, "
                        "a dictionary or keyword arguments."
                    )

        # Use kwargs to initialize OCI ModelDeployment attributes
        super().__init__(**oci_kwargs)

        # Get model ID from OCI object
        if self.model_id is None:
            try:
                self.model_id = (
                    self.model_deployment_configuration_details.model_configuration_details.model_id
                )
            except:
                self.model_id = None

        if self.model_id:
            self._config_model_id()

        # Process additional kwargs
        # Convert all keys to lower case
        kwargs = {str(k).lower(): v for k, v in kwargs.items()}

        # Set default display_name if not specified - randomly generated easy to remember name generated
        if "display_name" not in kwargs or kwargs["display_name"] is None:
            kwargs["display_name"] = utils.get_random_name_for_resource()

        # Set ModelDeployment properties
        for key, val in kwargs.items():
            if key in self.swagger_types.keys():
                self.with_prop(key, val)

        # Config instance
        instance_config = {}
        for key in [
            "instance_shape",
            "instance_count",
            "bandwidth_mbps",
            "memory_in_gbs",
            "ocpus",
        ]:
            if key in kwargs:
                instance_config[key] = kwargs[key]
        if instance_config:
            self.with_instance_configuration(config=instance_config)

        # Config logging
        if "access_log_group_id" in kwargs or "access_log_id" in kwargs:
            if not ("access_log_group_id" in kwargs and "access_log_id" in kwargs):
                raise ValueError(
                    "access_log_group_id and access_log_id must be specified at the same time."
                )
            self.with_access_log(kwargs["access_log_group_id"], kwargs["access_log_id"])

        if "predict_log_group_id" in kwargs or "predict_log_id" in kwargs:
            if not ("predict_log_group_id" in kwargs and "predict_log_id" in kwargs):
                raise ValueError(
                    "predict_log_group_id and predict_log_id must be specified at the same time."
                )
            self.with_predict_log(
                kwargs["predict_log_group_id"], kwargs["predict_log_id"]
            )

    def with_instance_configuration(self, config):
        """with_instance_configuration creates a ModelDeploymentDetails object with a specific config

        Parameters
        ----------
        config : dict
            dictionary containing instance configuration about the deployment.
            The following keys are supported:

            - instance_shape: str,
            - instance_count: int,
            - bandwidth_mbps: int,
            - memory_in_gbs: float,
            - ocpus: float

            The instance_shape and instance_count are required when creating a new deployment.
            They are optional when updating an existing deployment.

        Returns
        -------
        ModelDeploymentProperties
            self

        """
        # Convert all keys to lower case for backward compatibility
        config = {str(k).lower(): v for k, v in config.items()}

        single_type_model_deployment_object = (
            data_science_models.SingleModelDeploymentConfigurationDetails()
        )

        model_configuration_details_object = (
            data_science_models.ModelConfigurationDetails()
        )
        model_configuration_details_object.model_id = self.model_id

        # instance_configuration is required even though it can be initialized with empty values
        instance_configuration_object = data_science_models.InstanceConfiguration()
        if "instance_shape" in config:
            instance_configuration_object.instance_shape_name = config["instance_shape"]

        if config.get("memory_in_gbs") or config.get("ocpus"):
            instance_shape_config_details = (
                data_science_models.ModelDeploymentInstanceShapeConfigDetails(
                    memory_in_gbs=config.pop("memory_in_gbs", None),
                    ocpus=config.pop("ocpus", None),
                )
            )
            instance_configuration_object.model_deployment_instance_shape_config_details = (
                instance_shape_config_details
            )

        model_configuration_details_object.instance_configuration = (
            instance_configuration_object
        )

        # scaling_policy is required even though it can be initialized with empty values
        scaling_policy_object = data_science_models.FixedSizeScalingPolicy()
        if "instance_count" in config:
            scaling_policy_object.instance_count = int(config["instance_count"])
        model_configuration_details_object.scaling_policy = scaling_policy_object

        if "bandwidth_mbps" in config:
            model_configuration_details_object.bandwidth_mbps = config["bandwidth_mbps"]

        single_type_model_deployment_object.model_configuration_details = (
            model_configuration_details_object
        )
        self.model_deployment_configuration_details = (
            single_type_model_deployment_object
        )
        return self

    def with_category_log(self, log_type: str, group_id: str, log_id: str):
        """Adds category log configuration

        Parameters
        ----------
        log_type : str
            The type of logging to be configured. Must be "access" or "predict"
        group_id : str
            Log group ID of OCI logging service
        log_id : str
            Log ID of OCI logging service

        Returns
        -------
        ModelDeploymentProperties
            self

        Raises
        ------
        ValueError
            When log_type is invalid

        """
        if log_type not in ["access", "predict"]:
            raise ValueError(
                f'Invalid log type: {log_type}. Must be "access" or "predict".'
            )
        if not self.category_log_details:
            self.category_log_details = data_science_models.CategoryLogDetails()
        log_details = data_science_models.LogDetails()
        log_details.log_group_id = group_id
        log_details.log_id = log_id
        setattr(self.category_log_details, log_type, log_details)
        return self

    def with_access_log(self, log_group_id: str, log_id: str):
        """Adds access log config

        Parameters
        ----------
        group_id : str
            Log group ID of OCI logging service
        log_id : str
            Log ID of OCI logging service

        Returns
        -------
        ModelDeploymentProperties
            self

        """
        return self.with_category_log("access", log_group_id, log_id)

    def with_predict_log(self, log_group_id: str, log_id: str):
        """Adds predict log config

        Parameters
        ----------
        group_id : str
            Log group ID of OCI logging service
        log_id : str
            Log ID of OCI logging service

        Returns
        -------
        ModelDeploymentProperties
            self

        """
        return self.with_category_log("predict", log_group_id, log_id)

    def with_logging_configuration(
        self,
        access_log_group_id: str,
        access_log_id: str,
        predict_log_group_id: Optional[str] = None,
        predict_log_id: Optional[str] = None,
    ):
        """Adds OCI logging configurations for OCI logging service

        Parameters
        ----------
        access_log_group_id : str
            Log group ID of OCI logging service for access log
        access_log_id : str
            Log ID of OCI logging service for access log
        predict_log_group_id : str
            Log group ID of OCI logging service for predict log
        predict_log_id : str
            Log ID of OCI logging service for predict log

        Returns
        -------
        ModelDeploymentProperties
            self

        """
        self.with_access_log(access_log_group_id, access_log_id)
        if predict_log_group_id and predict_log_id:
            self.with_predict_log(predict_log_group_id, predict_log_id)
        return self

    def with_prop(self, property_name: str, value: Any):
        """Sets model deployment's `property_name` attribute to `value`

        Parameters
        ----------
        property_name : str
            Name of a model deployment property.
        value :
            New value for property attribute.

        Returns
        -------
        ModelDeploymentProperties
            self

        """
        if property_name not in self.swagger_types:
            raise AttributeError(
                f"OCI ModelDeployment does not have the attribute {property_name}"
            )
        setattr(self, property_name, value)
        return self

    def _config_model_id(self):
        """Sets the model ID in the model configuration details using the value of self.model_id
        This method should be called before preparing the payload for OCI API.
        """
        if (
            self.model_deployment_configuration_details
            and self.model_deployment_configuration_details.model_configuration_details
        ):
            self.model_deployment_configuration_details.model_configuration_details.model_id = (
                self.model_id
            )
        else:
            self.with_instance_configuration({})

    def to_oci_model(self, oci_model):
        """Convert properties into an OCI data model

        Parameters
        ----------
        oci_model : class
            The class of OCI data model, e.g., oci.data_science_models.CreateModelDeploymentDetails

        """
        if self.model_id:
            self._config_model_id()

        return super().to_oci_model(oci_model)

    def build(self) -> data_science_models.CreateModelDeploymentDetails:
        """Converts the deployment properties to OCI CreateModelDeploymentDetails object.
        Converts a model URI into a model OCID if user passed in a URI.

        Returns
        -------
        CreateModelDeploymentDetails
            A CreateModelDeploymentDetails instance ready for OCI API.

        """
        if self.model_uri:
            self.model_id = OCIClientManager().prepare_artifact(
                model_uri=self.model_uri,
                properties=dict(
                    display_name="testing",
                    project_id=self.project_id,
                    compartment_id=self.compartment_id,
                ),
            )
            self.model_deployment_configuration_details.model_configuration_details.model_id = (
                self.model_id
            )
        return self.to_oci_model(data_science_models.CreateModelDeploymentDetails)

    def to_update_deployment(self) -> data_science_models.UpdateModelDeploymentDetails:
        """Converts the deployment properties to OCI UpdateModelDeploymentDetails object.

        Returns
        -------
        CreateModelDeploymentDetails
            An UpdateModelDeploymentDetails instance ready for OCI API.

        """
        return self.to_oci_model(data_science_models.UpdateModelDeploymentDetails)

    def __eq__(self, other):
        return self.to_dict() == other.to_dict()



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
There are two potential candidates for oracle database driver -
* cx_Oracle
* oracledb

Preference is to use oracledb if it is available in the environment else choose cx_Oracle

If oracledb is loaded and user uses `wallet` to connect to database, prefer thick mode. Thin mode requires user to provide passphrase for PEM file which is not required in thick mode

If user uses DSN string copied from OCI console with OCI database setup for TLS connection, oracledb driver is preferred. If cx_Oracle is the only driver available, warn user that oracledb is preferred driver.
Note: We need to account for cx_Oracle though oracledb can operate in thick mode. The end user may be is using one of the old conda packs or an environment where cx_Oracle is the only available driver.
"""

from ads.common.utils import ORACLE_DEFAULT_PORT

import logging
import numpy as np
import os
import pandas as pd
import tempfile
from time import time
from typing import Dict, Optional, List, Union, Iterator
import zipfile
from ads.common.decorator.runtime_dependency import (
    OptionalDependency,
)

logger = logging.getLogger("ads.oracle_connector")
CX_ORACLE = "cx_Oracle"
PYTHON_ORACLEDB = "PYTHON_ORACLEDB"
PYTHON_DRIVER_NAME = None

try:
    import oracledb as oracle_driver  # Both the driver share same signature for the APIs that we are using.

    PYTHON_DRIVER_NAME = PYTHON_ORACLEDB
except:
    logger.info("oracledb package not found. Trying to load cx_Oracle")
    try:
        import cx_Oracle as oracle_driver

        PYTHON_DRIVER_NAME = CX_ORACLE
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            f"Neither `oracledb` nor `cx_Oracle` module was not found. Please run "
            f"`pip install {OptionalDependency.DATA}`."
        )


class OracleRDBMSConnection(oracle_driver.Connection):
    def __init__(
        self,
        user_name,
        password,
        service_name=None,
        wallet_file=None,
        sid=None,
        dsn=None,
        host=None,
        port=ORACLE_DEFAULT_PORT,
        **kwargs,
    ):
        logger.info(f"Using `{PYTHON_DRIVER_NAME}` for connection with Oracle database")
        if "wallet_location" in kwargs:
            wallet_file = kwargs.pop("wallet_location")
        if PYTHON_DRIVER_NAME == PYTHON_ORACLEDB and wallet_file:
            try:
                oracle_driver.init_oracle_client()
                logger.info(
                    "Running oracledb driver in thick mode. For mTLS based connection, thick mode is default."
                )
            except:
                logger.info(
                    "Could not use thick mode. The driver is running in thin mode. System might prompt for passphrase"
                )

        self.temp_dir = None
        self.tns_entries = {}
        kwargs["user"] = user_name
        kwargs["password"] = password
        if wallet_file:
            self._setup_wallet_file(wallet_file)
            if service_name:
                dsn = service_name
            if dsn in self.tns_entries:
                dsn = self.tns_entries[dsn]
                kwargs = kwargs.copy()
                kwargs["dsn"] = dsn
        elif dsn:
            kwargs["dsn"] = dsn
            logger.info("Using dsn for connection")
            logger.debug(f"DSN string is {dsn}")
            if PYTHON_DRIVER_NAME == CX_ORACLE and "protocol=tcps" in dsn:
                logger.warning(
                    "If you are connecting to Autonomous Database using TLS, install `oracledb` python package to use the connection string from OCI Autonomous Database Console."
                )
        elif service_name or sid:
            logger.info(f"Connecting to {host}:{port}/{service_name or sid}")
            if not host:
                raise ValueError(
                    "Missing `host` information or wallet file. To connect to the database without using wallet file, `host` infomration is required. Please set `host` or `wallet_file` with a valid value."
                )
            kwargs["dsn"] = self._construct_dsn_(host, port, service_name, sid)

        super().__init__(**kwargs)

        logger.info(
            f"RDBMS version: {'.'.join([str(x) for x in self.version.split('.')[:2]])}"
        )

    def __del__(self):
        if self.temp_dir is not None:
            self.temp_dir.cleanup()

    def _construct_dsn_(self, host, port, service_name=None, sid=None):
        if PYTHON_DRIVER_NAME == CX_ORACLE:
            return oracle_driver.makedsn(host, port, sid=sid, service_name=service_name)
        else:
            cp = oracle_driver.ConnectParams(
                host=host, port=port, service_name=service_name, sid=sid
            )
            return cp.get_connect_string()

    def _setup_wallet_file(self, wallet_file: str):
        # extract files in wallet zip to a temporary directory
        self.temp_dir = tempfile.TemporaryDirectory()
        zipfile.ZipFile(wallet_file).extractall(self.temp_dir.name)

        # parse tnsnames.ora to get list of entries and modify them to include the wallet location
        fname = os.path.join(self.temp_dir.name, "tnsnames.ora")
        for line in open(fname):
            pos = line.find(" = ")
            if pos < 0:
                continue
            name = line[:pos]
            entry = line[pos + 3 :].strip()
            key_phrase = "(security="
            pos = entry.find(key_phrase) + len(key_phrase)
            wallet_entry = f"(MY_WALLET_DIRECTORY={self.temp_dir.name})"
            entry = entry[:pos] + wallet_entry + entry[pos:]
            self.tns_entries[name] = entry

    def insert(
        self,
        table_name: str,
        df: pd.DataFrame,
        if_exists: str,
        batch_size=100000,
        encoding="utf-8",
    ):

        if if_exists not in ["fail", "replace", "append"]:
            raise ValueError(
                f"Unknown option `if_exists`={if_exists}. Valid options are 'fail', 'replace', 'append'"
            )
        start_time = time()

        df_orcl = df.copy()
        # "object" type can be actual objects, or just plain strings, when inserting into the
        # database we need to stringify these so they can be represented in a VARCHAR2 column

        # object_columns = df_orcl.select_dtypes(include=["object"]).columns
        # df_orcl = df_orcl.where(pd.notnull(df_orcl), None)
        # df_orcl[object_columns] = df_orcl[object_columns].astype(str)

        # prep column names for valid Oracle column names (alpha + # $ _)
        df_orcl.columns = df_orcl.columns.str.replace(r"\W+", "_", regex=True)
        table_exist = True
        with self.cursor() as cursor:

            if if_exists != "replace":
                try:
                    cursor.execute(f"SELECT 1 from {table_name} FETCH NEXT 1 ROWS ONLY")
                except Exception:
                    table_exist = False
                if if_exists == "fail" and table_exist:
                    raise ValueError(
                        f"Table {table_name} already exists. Set `if_exists`='replace' or 'append' to replace or append to the existing table"
                    )
            # Oracle doesn't have boolean so convert to 1/0
            df_orcl = df_orcl.where(
                df.applymap(type) != bool, df_orcl.replace({True: 1, False: 0})
            )

            type_mappings = {
                "bool": "NUMBER(1)",
                "int16": "INTEGER",
                "int32": "INTEGER",
                "int64": "INTEGER",
                "float16": "FLOAT",
                "float32": "FLOAT",
                "float64": "FLOAT",
                "datetime64": "TIMESTAMP",
            }

            # add in any string types as Oracle's VARCHAR type setting length to accommodate longest
            def get_max_str_len(df, column, encoding):
                return df[column].dropna().str.encode(encoding).map(len).max()

            longest_string_column = max(
                (
                    get_max_str_len(df_orcl, c, encoding)
                    for c in df_orcl.select_dtypes(
                        include=["object", "category"]
                    ).columns
                ),
                default=0,
            )

            logger.debug(f"Max string column value: {longest_string_column}")

            datatypes = {
                c: f"VARCHAR2({get_max_str_len(df_orcl, c, encoding)})"
                for c in df_orcl.select_dtypes(include=["object", "category"]).columns
            }

            for df_type, orcl_type in type_mappings.items():
                datatypes.update(
                    {
                        column: orcl_type
                        for column in df_orcl.select_dtypes(include=df_type).columns
                    }
                )

            if set(datatypes.keys()) != set(df_orcl.columns):
                raise Exception(
                    f"unable to determine oracle data type to use for column(s): {', '.join(set(df_orcl.columns)-set(datatypes.keys()))}"
                )

            # create table
            if if_exists == "replace":
                try:
                    cursor.execute(f"drop table {table_name}")
                except:
                    logger.info(f"Table {table_name} does not exist")
            if if_exists == "replace" or not table_exist:
                sql = (
                    f"create table {table_name} ("
                    + ", ".join([f"{col} {datatypes[col]}" for col in df_orcl.columns])
                    + ")"
                )
                logger.info(sql)
                try:
                    cursor.execute(sql)
                except Exception as e:
                    raise Exception(
                        f"'{e if e.args else 'Unexpected error encountered'}' with query: '{sql}'"
                    ) from e

            # insert
            bind_columns = ", ".join([f"{col}" for col in df_orcl.columns])
            bind_variables = ", ".join([f":{col}" for col in df_orcl.columns])
            sql = f"insert into {table_name}({bind_columns}) values({bind_variables})"

            logger.info(sql)

            # prevent buffer reallocation by locking in the longest string value
            # cursor.setinputsizes(None, longest_string_column)

            # replace NaN with None before turning into database records, important - don't
            # do this earlier in the logic because it can change the pandas
            # data types

            record_data = list(df_orcl.replace({np.nan: None}).itertuples(index=False))

            def chunks(lst: List, batch_size: int):
                """Yield successive batch_size chunks from lst."""
                for i in range(0, len(lst), batch_size):
                    yield lst[i : i + batch_size]

            for batch in chunks(record_data, batch_size=batch_size):

                cursor.executemany(sql, batch, batcherrors=True)

                for error in cursor.getbatcherrors():
                    logger.error(
                        f"Error: {error.message}, at row offset: {error.offset}"
                    )
                    raise RuntimeError(
                        f"Error: {error.message}, at row offset: {error.offset}"
                    )

            self.commit()

            duration = time() - start_time
            logger.info(
                f"inserted {df_orcl.shape[0]} rows at {df_orcl.shape[0]/duration:.2f} rows/seconds"
            )

    def _fetch_by_batch(self, cursor, chunksize):
        while True:
            rows = cursor.fetchmany(chunksize)
            if rows:
                yield rows
            else:
                break

    def query(
        self, sql: str, bind_variables: Optional[Dict], chunksize=None
    ) -> Union[pd.DataFrame, Iterator[pd.DataFrame]]:

        start_time = time()

        cursor = self.cursor()
        cursor.arraysize = 50000

        if chunksize:
            logger.info(f"Chunksize is {chunksize}")
            cursor.execute(sql, **bind_variables)
            columns = [row[0] for row in cursor.description]
            df = iter(
                (
                    pd.DataFrame(data=rows, columns=columns)
                    for rows in self._fetch_by_batch(cursor, chunksize)
                )
            )

        else:
            df = pd.DataFrame(
                cursor.execute(sql, **bind_variables),
                columns=[row[0] for row in cursor.description],
            )
            duration = time() - start_time
            logger.info(
                f"fetched {df.shape[0]} rows at {df.shape[0]/duration:.2f} rows/seconds"
            )

        return df



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import copy
import json
import os
import tempfile
from concurrent.futures import Future, ThreadPoolExecutor
from time import sleep
from typing import List, Dict

from ads.opctl import logger
from ads.opctl.backend.base import Backend
from ads.opctl.config.resolver import ConfigResolver
from ads.opctl.distributed.cmds import local_run, load_ini
from ads.opctl.constants import (
    ML_JOB_IMAGE,
    ML_JOB_GPU_IMAGE,
    DEFAULT_IMAGE_HOME_DIR,
    DEFAULT_IMAGE_SCRIPT_DIR,
    DEFAULT_IMAGE_CONDA_DIR,
    DEFAULT_NOTEBOOK_SESSION_CONDA_DIR,
    DEFAULT_NOTEBOOK_SESSION_SPARK_CONF_DIR,
)
from ads.opctl.utils import get_docker_client, is_in_notebook_session
from ads.opctl.utils import build_image, run_container, run_command
from ads.opctl.spark.cmds import (
    generate_core_site_properties_str,
    generate_core_site_properties,
)
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.pipeline.ads_pipeline import Pipeline, PipelineStep
from oci.data_science.models import PipelineStepRun


class CondaPackNotFound(Exception):
    pass


class LocalBackend(Backend):
    def __init__(self, config: Dict) -> None:
        """
        Initialize a LocalBackend object with given config.

        Parameters
        ----------
        config: dict
            dictionary of configurations
        """
        self.config = config

    def run(self):
        if self.config.get("version") == "v1.0":
            docker_image = self.config["spec"]["Infrastructure"]["spec"]["dockerImage"]
            # TODO: don't hard code api keys
            bind_volumes = {
                os.path.expanduser("~/.oci"): {
                    "bind": os.path.join(DEFAULT_IMAGE_HOME_DIR, ".oci"),
                    "mode": "ro",
                }
            }
            self._run_with_image_v1(bind_volumes)

        else:
            bind_volumes = {}
            if not is_in_notebook_session():
                bind_volumes = {
                    os.path.expanduser(
                        os.path.dirname(self.config["execution"]["oci_config"])
                    ): {"bind": os.path.join(DEFAULT_IMAGE_HOME_DIR, ".oci")}
                }
            if self.config["execution"].get("conda_slug", None):
                exit_code = self._run_with_conda_pack(bind_volumes)
            elif self.config["execution"].get("image"):
                exit_code = self._run_with_image(bind_volumes)
            else:
                raise ValueError("Either conda pack info or image should be specified.")

            if exit_code != 0:
                raise RuntimeError(
                    f"Job did not complete successfully. Exit code: {exit_code}. "
                    f"Run with the --debug argument to view container logs."
                )

    @runtime_dependency(module="docker", install_from=OptionalDependency.OPCTL)
    def init_vscode_container(self) -> None:
        """
        Create a .devcontainer.json file for development with VSCode.

        Returns
        -------
        None
        """
        source_folder = self.config["execution"].get("source_folder", None)
        if source_folder:
            source_folder = os.path.abspath(source_folder)
            if not os.path.exists(source_folder):
                raise FileNotFoundError(
                    f"source folder {source_folder} does not exist."
                )
            if self.config["execution"].get("gpu", False):
                image = self.config["execution"].get("image", ML_JOB_GPU_IMAGE)
            else:
                image = self.config["execution"].get("image", ML_JOB_IMAGE)
            oci_config_folder = os.path.expanduser(
                os.path.dirname(self.config["execution"]["oci_config"])
            )
            dev_container = {
                "image": image,
                "extensions": ["ms-python.python"],
                "mounts": [
                    f"source={oci_config_folder},target={os.path.join(DEFAULT_IMAGE_HOME_DIR, '.oci')},type=bind",
                ],
                "workspaceMount": f"source={source_folder},target={os.path.join(DEFAULT_IMAGE_HOME_DIR, os.path.basename(source_folder))},type=bind",
                "workspaceFolder": DEFAULT_IMAGE_HOME_DIR,
                "name": f"{image}-dev-env",
            }
            if image == ML_JOB_IMAGE or image == ML_JOB_GPU_IMAGE:
                conda_folder = os.path.expanduser(
                    self.config["execution"]["conda_pack_folder"]
                )
                dev_container["mounts"].append(
                    f"source={conda_folder},target={DEFAULT_IMAGE_CONDA_DIR},type=bind"
                )
                dev_container[
                    "postCreateCommand"
                ] = "conda init bash && source ~/.bashrc"
            else:
                raise ValueError(
                    "`--source-folder` option works with image `ml-job`, `ml-job-gpu` only. Those can be build with `ads opctl build-image`. Please check `ads opctl build-image -h`."
                )
        else:
            image = self.config["execution"].get("image", None)
            if not image:
                raise ValueError("Image must be specified.")
            else:
                dev_container = {
                    "image": image,
                    "mounts": [],
                    "extensions": ["ms-python.python"],
                }

        dev_container["containerEnv"] = self.config["execution"]["env_vars"]
        for k, v in self.config["execution"]["volumes"].items():
            dev_container["mounts"].append(
                f"source={os.path.abspath(k)},target={v['bind']},type=bind"
            )

        try:
            client = get_docker_client()
            client.api.inspect_image(image)
        except docker.errors.ImageNotFound:
            cmd = None
            if image == ML_JOB_IMAGE:
                cmd = "ads opctl build-image job-local"
            elif image == ML_JOB_GPU_IMAGE:
                cmd = "ads opctl build-image job-local -g"
            if cmd:
                raise RuntimeError(
                    f"Image {image} not found. Please run `{cmd}` to build the image."
                )
            else:
                raise RuntimeError(f"Image {image} not found.")
        if source_folder:
            with open(
                os.path.join(os.path.abspath(source_folder), ".devcontainer.json"), "w"
            ) as f:
                f.write(json.dumps(dev_container, indent=2))
            print(f"File {os.path.join(source_folder, '.devcontainer.json')} created.")
        else:
            with open(os.path.abspath(".devcontainer.json"), "w") as f:
                f.write(json.dumps(dev_container, indent=2))
            print(f"File {os.path.abspath('.devcontainer.json')} created.")

    def _run_with_conda_pack(self, bind_volumes: Dict) -> int:
        env_vars = self.config["execution"]["env_vars"]
        slug = self.config["execution"]["conda_slug"]
        image = self.config["execution"]["image"]

        # bind_volumes is modified in-place and does not need to be returned
        # it is returned just to be explicit that it is changed during this function call
        bind_volumes, env_vars = self._check_conda_pack_and_install_if_applicable(
            slug, bind_volumes, env_vars
        )
        bind_volumes = self._mount_source_folder_if_exists(bind_volumes)
        command = self._build_command_for_conda_run()
        if is_in_notebook_session():
            run_command(command, shell=True)
        else:
            conda_pack_path = os.path.join(
                os.path.expanduser(self.config["execution"]["conda_pack_folder"]), slug
            )
            if os.path.exists(os.path.join(conda_pack_path, "spark-defaults.conf")):
                env_vars["SPARK_CONF_DIR"] = os.path.join(DEFAULT_IMAGE_CONDA_DIR, slug)
            return self._activate_conda_env_and_run(
                image, slug, command, bind_volumes, env_vars
            )

    def _build_command_for_conda_run(self) -> str:
        if ConfigResolver(self.config)._is_ads_operator():
            if is_in_notebook_session():
                curr_dir = os.path.dirname(os.path.abspath(__file__))
                script = os.path.abspath(
                    os.path.join(curr_dir, "..", "operators", "run.py")
                )
            else:
                script = os.path.join(DEFAULT_IMAGE_SCRIPT_DIR, "operators/run.py")
            command = f"python {script} "
            if self.config["execution"]["auth"] == "resource_principal":
                command += "-r "
        else:
            entry_script = self.config["execution"].get("entrypoint")
            if not entry_script:
                raise ValueError(
                    "An entrypoint script must be specified when running with conda pack. "
                    "Use `--entrypoint`."
                )
            if not os.path.exists(
                os.path.join(self.config["execution"]["source_folder"], entry_script)
            ):
                raise FileNotFoundError(
                    f"{os.path.join(self.config['execution']['source_folder'], entry_script)} is not found."
                )
            if is_in_notebook_session():
                source_folder = os.path.join(self.config["execution"]["source_folder"])
            else:
                source_folder = os.path.join(
                    DEFAULT_IMAGE_SCRIPT_DIR,
                    "operators",
                    os.path.basename(self.config["execution"]["source_folder"]),
                )
            if os.path.splitext(entry_script)[-1] == ".py":
                command = f"python {os.path.join(source_folder, entry_script)} "
                if is_in_notebook_session():
                    command = (
                        f"source activate {os.path.join(DEFAULT_NOTEBOOK_SESSION_CONDA_DIR, self.config['execution']['conda_slug'])} && "
                        + command
                    )
            elif os.path.splitext(entry_script)[-1] == ".sh":
                command = f"cd {source_folder} && /bin/bash {entry_script} "
            else:
                logger.warn(
                    "ML Job only support .py and .sh files."
                    "If you intend to submit to ML Job later, please update file extension."
                )
                command = f"cd {source_folder} && {entry_script} "
        if self.config["execution"].get("command"):
            command += f"{self.config['execution']['command']}"
        return command

    def _run_with_image(self, bind_volumes: Dict) -> int:
        env_vars = self.config["execution"]["env_vars"]
        image = self.config["execution"]["image"]
        if ConfigResolver(self.config)._is_ads_operator():
            # running operators
            command = (
                f"python {os.path.join(DEFAULT_IMAGE_SCRIPT_DIR, 'operators/run.py')} "
            )
            entrypoint = None
            if self.config["execution"].get("command", None):
                command += f"{self.config['execution']['command']}"
        else:
            # in case of running a user image, entrypoint is not required
            entrypoint = self.config["execution"].get("entrypoint", None)
            command = self.config["execution"].get("command", None)
        if self.config["execution"].get("source_folder", None):
            bind_volumes.update(self._mount_source_folder_if_exists(bind_volumes))
        bind_volumes.update(self.config["execution"]["volumes"])
        return run_container(image, bind_volumes, env_vars, command, entrypoint)

    def _run_with_image_v1(self, bind_volumes: Dict) -> int:
        env_vars = [
            str(d["name"]) + "=" + str(d["value"])
            for d in self.config["spec"]["Runtime"]["spec"]["environmentVariables"]
        ]
        image = self.config["spec"]["Infrastructure"]["spec"]["dockerImage"]
        command = self.config["spec"]["Runtime"]["spec"]["entrypoint"]
        entrypoint = "python /etc/datascience/operator/cluster_helper.py"

        print("looking to bind volume")
        bind_volumes.update(self.config["spec"]["Framework"]["spec"]["bindVolumes"])
        return run_container(
            image=image,
            bind_volumes=bind_volumes,
            env_vars=env_vars,
            command=command,
            entrypoint=entrypoint,
        )

    def _check_conda_pack_and_install_if_applicable(
        self, slug: str, bind_volumes: Dict, env_vars: Dict
    ) -> Dict:
        conda_pack_path = os.path.join(
            os.path.expanduser(self.config["execution"]["conda_pack_folder"]), slug
        )
        if not os.path.exists(conda_pack_path):
            raise CondaPackNotFound(
                f"Conda pack {conda_pack_path} not found. Please run `ads opctl conda create` or `ads opctl conda install`."
            )
        if os.path.exists(os.path.join(conda_pack_path, "spark-defaults.conf")):
            if not is_in_notebook_session():
                env_vars["SPARK_CONF_DIR"] = os.path.join(DEFAULT_IMAGE_CONDA_DIR, slug)
            # write core_site.xml
            if self.config["execution"]["auth"] == "api_key":
                properties = generate_core_site_properties(
                    "api_key",
                    self.config["execution"]["oci_config"],
                    self.config["execution"]["oci_profile"],
                )
                # key path cannot have "~/"
                oci_config_folder = os.path.expanduser(
                    os.path.dirname(self.config["execution"]["oci_config"])
                )
                properties[-1] = (
                    properties[-1][0],
                    os.path.join(
                        DEFAULT_IMAGE_HOME_DIR,
                        ".oci",
                        os.path.relpath(
                            os.path.expanduser(properties[-1][1]), oci_config_folder
                        ),
                    ),
                )
            else:
                properties = generate_core_site_properties("resource_principal")

            core_site_str = generate_core_site_properties_str(properties)
            if is_in_notebook_session():
                with open(
                    os.path.join(
                        DEFAULT_NOTEBOOK_SESSION_SPARK_CONF_DIR, "core-site.xml"
                    ),
                    "w",
                ) as f:
                    f.write(core_site_str)
            else:
                with open(os.path.join(conda_pack_path, "core-site.xml"), "w") as f:
                    f.write(core_site_str)
        bind_volumes[
            os.path.abspath(
                os.path.expanduser(
                    os.path.join(self.config["execution"]["conda_pack_folder"], slug)
                )
            )
        ] = {"bind": os.path.join(DEFAULT_IMAGE_CONDA_DIR, slug)}
        return bind_volumes, env_vars

    def _mount_source_folder_if_exists(self, bind_volumes: Dict) -> Dict:
        source_folder = os.path.abspath(self.config["execution"]["source_folder"])
        if not os.path.exists(source_folder):
            raise FileNotFoundError(f"source folder {source_folder} does not exist.")
        mount_path = os.path.join(
            DEFAULT_IMAGE_SCRIPT_DIR,
            "operators",
            os.path.basename(self.config["execution"]["source_folder"]),
        )
        bind_volumes[source_folder] = {"bind": mount_path}
        return bind_volumes

    @staticmethod
    @runtime_dependency(module="docker", install_from=OptionalDependency.OPCTL)
    def _activate_conda_env_and_run(
        image: str, slug: str, command: List[str], bind_volumes: Dict, env_vars: Dict
    ) -> int:
        try:
            client = get_docker_client()
            client.api.inspect_image(image)
        except docker.errors.ImageNotFound:
            logger.info(f"Image {image} not found. Attempt building it now....")
            if image == ML_JOB_IMAGE:
                build_image("job-local", gpu=False)
            else:
                build_image("job-local", gpu=True)

        with tempfile.TemporaryDirectory() as td:
            with open(os.path.join(td, "entryscript.sh"), "w") as f:
                f.write(
                    f"""
#!/bin/bash
source {os.path.join(DEFAULT_IMAGE_CONDA_DIR, slug, 'bin/activate')}
{command}
                        """
                )
            bind_volumes[os.path.join(td, "entryscript.sh")] = {
                "bind": os.path.join(DEFAULT_IMAGE_SCRIPT_DIR, "entryscript.sh")
            }
            env_vars["conda_slug"] = slug
            return run_container(
                image,
                bind_volumes,
                env_vars,
                command=f"bash {os.path.join(DEFAULT_IMAGE_SCRIPT_DIR, 'entryscript.sh')}",
            )


class LocalBackendDistributed(LocalBackend):
    def __init__(self, config: Dict) -> None:
        """
        Initialize a LocalBackendDistributed object with given config. This serves local single node(docker) testing
        for Distributed Tranining


        Parameters
        ----------
        config: dict
            dictionary of configurations
        """
        self.config = config

    def run(self):
        local_run(self.config, load_ini())


class LocalPipelineBackend(Backend):
    LOG_PREFIX = "Local Pipeline:"
    DEFAULT_PARALLEL_CONTAINER_MAXIMUM = 4
    DEFAULT_STATUS_POLL_INTERVAL_SECONDS = 5

    def __init__(self, config: Dict) -> None:
        """
        Initialize a LocalPipelineBackend object with given config.

        Parameters
        ----------
        config: dict
            dictionary of configurations
        """
        self.config = config

    def run(self) -> None:
        pipeline = Pipeline.from_dict(self.config)

        self._log_orchestration_message(f"Starting pipeline {pipeline.name} locally.")
        if pipeline.dag:
            self._log_orchestration_message(f"Pipeline DAG:")
            for d in pipeline.dag:
                self._log_orchestration_message(f"  {d}")

        completed_status = {}
        waiting_steps = {}
        for s in pipeline.step_details:
            waiting_steps[s.name] = s
        futures = {}
        pipeline_failure = False
        done = False

        if "max_parallel_containers" in self.config["infrastructure"]:
            max_parallel_containers = int(
                self.config["infrastructure"]["max_parallel_containers"]
            )
        else:
            max_parallel_containers = min(
                self.DEFAULT_PARALLEL_CONTAINER_MAXIMUM, os.cpu_count()
            )
            logger.warn(
                f"max_parallel_containers not specified in the config. Defaulting to {max_parallel_containers}."
                " Run `ads opctl configure` to define your local backend config."
            )

        poll_interval_seconds = int(
            self.config["infrastructure"].get(
                "pipeline_status_poll_interval_seconds",
                self.DEFAULT_STATUS_POLL_INTERVAL_SECONDS,
            )
        )

        with ThreadPoolExecutor(max_workers=max_parallel_containers) as executor:
            while not done:
                # Check if any running steps have completed
                for s in list(futures):
                    if futures[s].done():
                        if futures[s].exception() is None:
                            self._log_orchestration_message(
                                f"Step {s} completed successfully."
                            )
                            completed_status[
                                s
                            ] = PipelineStepRun.LIFECYCLE_STATE_SUCCEEDED
                        else:
                            pipeline_failure = True
                            self._log_orchestration_message(f"Step {s} failed:")
                            logger.error(futures[s].exception())
                            completed_status[s] = PipelineStepRun.LIFECYCLE_STATE_FAILED
                        del futures[s]

                for s in list(waiting_steps):
                    # Cancel all waiting steps if a failure is encountered
                    if pipeline_failure:
                        self._log_orchestration_message(
                            f"Skipping step {s} - pipeline failure encountered."
                        )
                        completed_status[s] = PipelineStepRun.LIFECYCLE_STATE_SKIPPED
                        del waiting_steps[s]
                        continue

                    # Start a waiting step if all of its dependencies have completed successfully
                    completed_deps = [
                        dep
                        for dep in waiting_steps[s].depends_on
                        if dep in completed_status
                    ]
                    if len(waiting_steps[s].depends_on) == len(completed_deps):
                        self._log_orchestration_message(f"Starting step {s}")
                        futures[s] = self._start_pipeline_step(
                            waiting_steps[s], executor
                        )
                        del waiting_steps[s]

                if len(completed_status) == len(pipeline.step_details):
                    done = True
                else:
                    sleep(poll_interval_seconds)

        self._log_orchestration_message("Pipeline run complete!")
        self._log_orchestration_message("Summary:")
        for step in pipeline.step_details:
            self._log_orchestration_message(
                f"  {step.name} - {completed_status[step.name]}"
            )

    def _start_pipeline_step(
        self, step: PipelineStep, executor: ThreadPoolExecutor
    ) -> Future:
        """
        Starts a single pipeline step.

        Parameters
        ----------
        step: PipelineStep
            The pipeline step to start
        executor: ThreadPoolExecutor
            The executor used to run the pipeline step.

        Returns
        -------
        future: Future
            The Future that can be used to query the status of the pipeline step.
        """
        step_config = self._create_step_config(step)

        # Have a local job backend execute the step using the updated step config
        local_job = LocalBackend(step_config)
        return executor.submit(local_job.run)

    def _create_step_config(self, pipeline_step: PipelineStep) -> Dict:
        """
        Creates the config for local execution of an individual pipeline step.

        Parameters
        ----------
        pipeline_step: PipelineStep
            The pipeline step whose config should be generated

        Returns
        -------
        step_config: Dict
            The config for the individual pipeline step.
        """
        if pipeline_step.kind.upper() != "CUSTOM_SCRIPT":
            raise ValueError(
                f"Step {pipeline_step.name} has unsupported kind. "
                f"Local pipeline execution only supports pipeline steps with kind customScript."
            )

        step_config = copy.deepcopy(self.config)
        step_config["kind"] = pipeline_step.kind
        step_config["type"] = pipeline_step.type
        del step_config["spec"]
        step_execution_config = step_config["execution"]
        step_execution_config["conda_slug"] = pipeline_step.runtime.conda["slug"]
        step_execution_config["env_vars"] = pipeline_step.runtime.envs

        if pipeline_step.runtime.type == "script":
            step_execution_config["entrypoint"] = pipeline_step.runtime.script_uri
        elif pipeline_step.runtime.type == "python":
            step_execution_config["entrypoint"] = pipeline_step.runtime.script_uri
            step_execution_config["source_folder"] = pipeline_step.runtime.working_dir
        elif pipeline_step.runtime.type == "notebook":
            step_execution_config["entrypoint"] = pipeline_step.runtime.notebook_uri
        else:
            raise ValueError(
                f"Step {pipeline_step.name} has unsupported runtime. "
                f"Supported values are: script, python, notebook"
            )

        if not step_execution_config.get("source_folder"):
            logger.warn(
                "No source_folder provided; defaulting to the current working directory. To specify a source"
                "folder for all pipeline steps, use the --source-folder parameter. To specify a source folder"
                "for individual steps, use a runtime with type python and specify the workingDir property."
            )
            step_execution_config["source_folder"] = os.getcwd()

        ConfigResolver(step_config).process()
        return step_config

    def _log_orchestration_message(self, str: str) -> None:
        """
        Logs a message related to pipeline run orchestration

        Parameters
        ----------
        str: str
            The message to log
        """
        logger.info(f"{self.LOG_PREFIX}: {str}")



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from __future__ import print_function, absolute_import

import pandas as pd

from ads.common import utils
from ads.type_discovery.constant_detector import ConstantDetector
from ads.type_discovery.continuous_detector import ContinuousDetector
from ads.type_discovery.credit_card_detector import CreditCardDetector
from ads.type_discovery.datetime_detector import DateTimeDetector
from ads.type_discovery.discrete_detector import DiscreteDetector
from ads.type_discovery.document_detector import DocumentDetector
from ads.type_discovery.latlon_detector import LatLonDetector
from ads.type_discovery.phone_number_detector import PhoneNumberDetector
from ads.type_discovery.unknown_detector import UnknownDetector
from ads.type_discovery.zipcode_detector import ZipCodeDetector
from ads.type_discovery.typed_feature import ContinuousTypedFeature

#
# these should be in precendence order, from low to high
#
discovery_plugins = [
    ConstantDetector(),
    DocumentDetector(),
    ZipCodeDetector(),
    LatLonDetector(),
    CreditCardDetector(),
    PhoneNumberDetector(),
    DateTimeDetector(),
    DiscreteDetector(),
    ContinuousDetector(),
    UnknownDetector(),
]


class TypeDiscoveryDriver:

    #
    # takes a pandas series
    #
    def discover(self, name: str, s: pd.Series, is_target: bool = False):
        """return the type of series

        Parameters
        ----------
        name : type
            variable name to discover.
        s : type
            series of values to 'type'
        is_target : type
            when true the rules differ, any continuous is contunuous regardless of other rules

        Returns
        -------
        type
            one of:

                ConstantDetector,
                DocumentDetector,
                ZipCodeDetector,
                LatLonDetector,
                CreditCardDetector,
                PhoneNumberDetector,
                DateTimeDetector,
                DiscreteDetector,
                ContinuousDetector,

        """

        assert (
            type(s).__name__ == "Series"
        ), "Type discovery can only be performed on a pandas.Series"

        if is_target and ContinuousDetector._target_is_continuous(s):
            return ContinuousTypedFeature.build(name, s)

        #
        # to lazily evaluate the discover method we use a generator expression iterable
        #
        return utils.first_not_none(
            (plugin.discover(name, s) for plugin in discovery_plugins)
        )



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


import os
import json

from oci.data_flow.models import Application

from ads.opctl.config.utils import read_from_ini
from ads.opctl.constants import (
    ADS_DATAFLOW_CONFIG_FILE_NAME,
    DEFAULT_ADS_CONFIG_FOLDER,
)

from ads.jobs import logger
from ads.common.utils import oci_key_profile


def get_dataflow_config(path=None, oci_profile=None):
    if path:
        dataflow_config_file_path = os.path.abspath(os.path.expanduser(path))
    else:
        dataflow_config_file_path = os.path.expanduser(
            os.path.join(DEFAULT_ADS_CONFIG_FOLDER, ADS_DATAFLOW_CONFIG_FILE_NAME)
        )
    config = {}
    if os.path.exists(dataflow_config_file_path):
        parser = read_from_ini(dataflow_config_file_path)
        if not oci_profile:
            oci_profile = oci_key_profile()
        if oci_profile in parser:
            config = dict(parser[oci_profile])
        if len(config) == 0:
            logger.error(
                f"Dataflow configuration with profile {oci_profile} not found."
            )
            raise ValueError(
                f"Dataflow configuration with profile {oci_profile} not found."
            )
        return config
    else:
        logger.error(f"{dataflow_config_file_path} not found.")
        return {}


class DataFlowConfig(Application):
    def __init__(self, path: str = None, oci_profile: str = None):
        """Create a DataFlowConfig object. If a path to config file is given it is loaded from the path.

        Parameters
        ----------
        path : str, optional
            path to configuration file, by default None
        oci_profile : str, optional
            oci profile to use, by default None
        """
        self.config = get_dataflow_config(path, oci_profile)
        self._script_bucket = None
        self._archive_bucket = None
        if len(self.config) > 0:
            self._script_bucket = self.config.pop("script_bucket")
            self._archive_bucket = self.config.pop("archive_bucket", None)
        super().__init__(**self.config)

    def __repr__(self):
        config = json.loads(super().__repr__())
        config["script_bucket"] = self.script_bucket
        if self.archive_bucket:
            config["archive_bucket"] = self.archive_bucket
        return f"'{json.dumps({k: v for k, v in config.items() if v is not None})}'"

    @property
    def script_bucket(self):
        """Bucket to save user script. Also accept a prefix in the format of oci://<bucket-name>@<namespace>/<prefix>.

        Returns
        -------
        str
            script bucket (path)
        """
        return self._script_bucket

    @script_bucket.setter
    def script_bucket(self, v: str):
        self._script_bucket = v

    @property
    def archive_bucket(self):
        """Bucket to save archive zip. Also accept a prefix in the format of oci://<bucket-name>@<namespace>/<prefix>.

        Returns
        -------
        str :
            archive bucket (path)
        """
        return self._archive_bucket

    @archive_bucket.setter
    def archive_bucket(self, v: str):
        self._archive_bucket = v



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import warnings
from enum import Enum
from functools import wraps


class NotSupportedError(Exception):
    pass


class TARGET_TYPE(Enum):
    CLASS = "Class"
    METHOD = "Method"
    ATTRIBUTE = "Attribute"


def deprecated(
    deprecated_in: str = None,
    removed_in: str = None,
    details: str = None,
    target_type: str = None,
    raise_error: bool = False,
):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.

    Parameters
    ----------
    deprecated_in: `str`
        Version of ADS where this function deprecated.
    removed_in: `str`
        Future version where this function will be removed.
    details: `str`
        More information to be shown.
    """

    def decorator(target):
        @wraps(target)
        def wrapper(*args, **kwargs):
            if target.__name__ == "__init__":
                _target_type = target_type or TARGET_TYPE.CLASS.value
                target_name = target.__qualname__.split(".")[0]
            else:
                _target_type = target_type or TARGET_TYPE.METHOD.value
                target_name = target.__name__
            if deprecated_in:
                msg = (
                    f"{_target_type} {target_name} is "
                    f"deprecated in {deprecated_in} and will be "
                    f"removed in {removed_in if removed_in else 'a future release'}."
                    f"{'' if not details else ' ' + details}"
                )
            else:
                msg = details
            if raise_error:
                raise NotSupportedError(msg)
            else:
                warnings.warn(msg, DeprecationWarning, stacklevel=2)
            return target(*args, **kwargs)

        return wrapper

    return decorator



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


class NTrials:
    """
    Exit based on number of trials.

    Parameters
    ----------
    n_trials: int
        Number of trials (sets of hyperparamters tested). If :obj:`None`, there is no
        limitation on the number of trials.

    Returns
    -------
        :class:`NTrials`
            NTrials object
    """

    def __init__(self, n_trials: int):
        self.n_trials = n_trials

    def __call__(self):
        return self.n_trials


class TimeBudget:
    """
    Exit based on the number of seconds.

    Parameters
    ----------
    seconds: float
        Time limit, in seconds. If :obj:`None` there is no time limit.

    Returns
    -------
        :class:`TimeBudget`
            TimeBudget object
    """

    def __init__(self, seconds: float):
        assert seconds > 0, "<code>time_budget</code> has to be greater than 0."
        self.seconds = seconds

    def __call__(self):
        return self.seconds


class ScoreValue:
    """
    Exit if the score is greater than or equal to the threshold.

    Parameters
    ----------
    score: float
        The threshold for exiting the tuning process. If a trial value is greater or equal
        to `score`, process exits.

    Returns
    -------
        :class:`ScoreValue`
            ScoreValue object
    """

    def __init__(self, score: float):
        self.score = score

    def __call__(self, study: "optuna.study.Study", trial: "optuna.trial.FrozenTrial"):
        if trial.value >= self.score:
            study.stop()



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2020, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import logging
from typing import List
from numpy.typing import ArrayLike

logger = logging.getLogger(__name__)


class EvaluatorMixin:
    def evaluate(
        self,
        X: ArrayLike,
        y: ArrayLike,
        y_pred: ArrayLike = None,
        y_score: ArrayLike = None,
        X_train: ArrayLike = None,
        y_train: ArrayLike = None,
        classes: List = None,
        positive_class: str = None,
        legend_labels: dict = None,
        perfect: bool = True,
        filename: str = None,
        use_case_type: str = None,
    ):
        """Creates an ads evaluation report.

        Parameters
        ----------
        X : DataFrame-like
            The data used to make a prediction.
            Can be set to None if `y_preds` is given. (And `y_scores` for more thorough analysis).
        y : array-like
            The true values corresponding to the input data
        y_pred : array-like, optional
            The predictions from each model in the same order as the models
        y_score : array-like, optional
            The predict_probas from each model in the same order as the models
        X_train : DataFrame-like, optional
            The data used to train the model
        y_train : array-like, optional
            The true values corresponding to the input training data
        classes : List or None, optional
            A List of the possible labels for y, when evaluating a classification use case
        positive_class : str or int, optional
            The class to report metrics for binary dataset. If the target classes is True or False,
            positive_class will be set to True by default. If the dataset is multiclass or multilabel,
            this will be ignored.
        legend_labels : dict, optional
            List of legend labels. Defaults to `None`.
            If legend_labels not specified class names will be used for plots.
        use_case_type : str, optional
            The type of problem this model is solving. This can be set during `prepare()`.
            Examples: "binary_classification", "regression", "multinomial_classification"
            Full list of supported types can be found here: `ads.common.model_metadata.UseCaseType`
        filename: str, optional
            If filename is given, the html report will be saved to the location specified.

        Examples
        --------
        >>> import tempfile
        >>> from ads.evaluations.evaluator import Evaluator
        >>> from sklearn.tree import DecisionTreeClassifier
        >>> from sklearn.datasets import make_classification
        >>> from sklearn.model_selection import train_test_split
        >>> from ads.model.framework.sklearn_model import SklearnModel
        >>> from ads.common.model_metadata import UseCaseType
        >>>
        >>> X, y = make_classification(n_samples=1000)
        >>> X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.3)
        >>> est = DecisionTreeClassifier().fit(X_train, y_train)
        >>> model = SklearnModel(estimator=est, artifact_dir=tempfile.mkdtemp())
        >>> model.prepare(
                inference_conda_env="generalml_p38_cpu_v1",
                training_conda_env="generalml_p38_cpu_v1",
                X_sample=X_test,
                y_sample=y_test,
                use_case_type=UseCaseType.BINARY_CLASSIFICATION,
            )
        >>> model.evaluate(X_test, y_test, filename="report.html")
        """

        from ads.evaluations.evaluator import Evaluator

        y_preds, y_scores = (
            ([y_pred], [y_score]) if y_pred is not None else (None, None)
        )
        report = Evaluator(
            models=[self],
            X=X,
            y=y,
            y_preds=y_preds,
            y_scores=y_scores,
            X_train=X_train,
            y_train=y_train,
            classes=classes,
            positive_class=positive_class,
            legend_labels=legend_labels,
            use_case_type=use_case_type,
        )
        if filename is None:
            report.display(perfect=perfect)
        else:
            report.save(filename=filename, perfect=perfect)
        return report



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from dataclasses import dataclass
from os import environ
from typing import Union

from ads.model.base_properties import BaseProperties


@dataclass(repr=False)
class ModelProperties(BaseProperties):
    """Represents properties required to save and deploy model."""

    inference_conda_env: str = None
    inference_python_version: str = None
    training_conda_env: str = None
    training_python_version: str = None
    training_resource_id: str = None
    training_script_path: str = None
    training_id: str = None
    compartment_id: str = None
    project_id: str = None
    bucket_uri: str = None
    remove_existing_artifact: bool = None
    overwrite_existing_artifact: bool = None
    deployment_instance_shape: str = None
    deployment_instance_count: int = None
    deployment_bandwidth_mbps: int = None
    deployment_log_group_id: str = None
    deployment_access_log_id: str = None
    deployment_predict_log_id: str = None
    deployment_memory_in_gbs: Union[float, int] = None
    deployment_ocpus: Union[float, int] = None

    def _adjust_with_env(self) -> None:
        """Adjusts env variables. This method is used within `with_env` method."""

        super()._adjust_with_env()
        props_env_map = {
            "project_id": ["PROJECT_OCID"],
            "training_resource_id": ["JOB_RUN_OCID", "NB_SESSION_OCID"],
            "compartment_id": [
                "JOB_RUN_COMPARTMENT_OCID",
                "NB_SESSION_COMPARTMENT_OCID",
            ],
        }
        for key, env_keys in props_env_map.items():
            try:
                value = next(
                    environ.get(env_key)
                    for env_key in env_keys
                    if environ.get(env_key, None) is not None
                )
                self.with_prop(key, value)
            except:
                pass
